# Generated by Django 3.2.13 on 2022-06-01 22:17

from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.db import migrations
from vendor.models import Payment


def post_payment_status_change(apps, schema_editor):
    ReceiptModel = apps.get_model('vendor', 'Receipt')
    PaymentModel = apps.get_model('vendor', 'Payment')

    for payment in PaymentModel.objects.all():
        try:    
            receipt = ReceiptModel.objects.get(transaction=payment.transaction, created__year=payment.created.year, created__month=payment.created.month, created__day=payment.created.day)
        except ObjectDoesNotExist:
            receipt = None
        except MultipleObjectsReturned:
            receipt = ReceiptModel.objects.filter(transaction=payment.transaction, created__year=payment.created.year, created__month=payment.created.month, created__day=payment.created.day).first()
        
        if receipt:
            r = ReceiptModel.objects.get(pk=receipt.pk)
            payment.status = r.status
            payment.save()
            print(f"Payment Status Updated {payment}")


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0035_add_subscription_relation'),
    ]

    operations = [
        migrations.RunPython(post_payment_status_change, reverse_code=migrations.RunPython.noop),
    ]




# Generated by Django 3.2.19 on 2023-07-12 16:40

from django.db import migrations


def rename_stripe_processor_site_settings(apps, schema_editor):
    SiteConfigModel = apps.get_model('siteconfigs', 'SiteConfigModel')

    for site_setting in SiteConfigModel.objects.filter(key='vendor.config.PaymentProcessorSiteConfig', value__payment_processor='stripe_processor.StripeProcessor'):
        site_setting.value['payment_processor'] = 'stripe.StripeProcessor'
        site_setting.save()
        print(f"\nStripe Payment Processor Config Update for site {site_setting.site}")


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0043_invoice_global_discount'),
    ]

    operations = [
        migrations.RunPython(rename_stripe_processor_site_settings, reverse_code=migrations.RunPython.noop)
    ]




from django import template
from django.urls import reverse
from vendor.config import PaymentProcessorSiteConfig, StripeConnectAccountConfig, VendorSiteCommissionConfig
from siteconfigs.models import SiteConfigModel

register = template.Library()

@register.inclusion_tag('vendor/includes/edit_link.html')
def config_edit_link(config_key, config_pk):
    config = SiteConfigModel.objects.get(pk=config_pk)
    link = ''

    if config.key == PaymentProcessorSiteConfig().key:
        link = reverse('vendor_admin:manager-config-processor-update', kwargs={'pk': config_pk})
    elif config.key == StripeConnectAccountConfig().key:
        link = reverse('vendor_admin:manager-config-stripe-connect-update', kwargs={'pk': config_pk})
    elif config.key == VendorSiteCommissionConfig().key:
        link = reverse('vendor_admin:manager-config-commission-update', kwargs={'pk': config_pk})

    return {
        'link': link
    }


import logging
from datetime import datetime

from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.utils import timezone

from vendor.models import CustomerProfile, Offer, Subscription
from vendor.models.choice import InvoiceStatus, SubscriptionStatus
from vendor.processors import AuthorizeNetProcessor, StripeProcessor

logger = logging.getLogger(__name__)


def convert_authorize_net_submitted_date(submitted_date):
    try:
        submitted_datetime = datetime.strptime(submitted_date.pyval, '%Y-%m-%dT%H:%M:%S.%f%z')
    
    except ValueError as exce:
        logger.error(f"sync_subscriptions_and_create_missing_receipts error {exce}")
        submitted_datetime = datetime.strptime(submitted_date.pyval, '%Y-%m-%dT%H:%M:%S%z')

    return submitted_datetime


def get_last_transaction_datetime(transactions, authorize_net):
    if not transactions:
        return None
    
    if len(transactions) == 1:
        transaction_id = transactions[0].transId.text
        transaction_detail = authorize_net.get_transaction_detail(transaction_id)
        return convert_authorize_net_submitted_date(transaction_detail.submitTimeUTC)
    
    return convert_authorize_net_submitted_date(transactions[0].submitTimeUTC)

def transfer_subscriptions_from_authorizenet_to_stripe(site):
    authorize_net = AuthorizeNetProcessor(site)
    stripe = StripeProcessor(site)

    auth_subscriptions = authorize_net.get_list_of_subscriptions()

    for auth_subscription in auth_subscriptions:
        subscription_info = authorize_net.subscription_info(auth_subscription.id.text)
        logger.info(f"Syncing AuthorizeNet Subscription ID: {auth_subscription.id.text}")
        subscription_transactions = authorize_net.get_subscription_transactions(subscription_info)
        last_transaction_datetime = get_last_transaction_datetime(subscription_transactions, authorize_net)

        if hasattr(subscription_info.subscription.profile, 'email'):
            email = subscription_info.subscription.profile.email.text

            try:
                customer_profile = CustomerProfile.objects.get(site=site, user__email__iexact=email)
                offers = Offer.objects.filter(site=site, name__contains=subscription_info.subscription.name)
                stripe_customer = stripe.get_stripe_customer_from_email(email)
                
                if not stripe_customer:
                    stripe.create_stripe_customers([customer_profile])
                elif stripe_customer and 'stripe_id' not in customer_profile.meta:
                    customer_profile.meta.update({'stripe_id': stripe_customer.id})
                    customer_profile.save()

                stripe_customer = stripe.stripe_get_object(stripe.stripe.Customer, customer_profile.meta['stripe_id'], expand=['subscriptions'])
                
                if 'site' not in stripe_customer['metadata']:
                    stripe.update_stripe_customers([customer_profile])

                if not offers.count():
                    raise ObjectDoesNotExist()
                offer = offers.first()
                
                invoice = customer_profile.get_cart()
                invoice.empty_cart()
                invoice.add_offer(offer)
                stripe.invoice = invoice

                for stripe_customer_subscription in stripe_customer.subscriptions:
                    stripe_product = stripe.stripe_get_object(stripe.stripe.Product, stripe_customer_subscription['plan']['product'])
                    if stripe_product['metadata']['pk'] == str(offer.pk) and 'stripe' not in offer.meta:
                        offer.meta.update({'stripe': {'product_id': stripe_product.id}})
                        offer.save()
                        stripe.sync_offer(offer)
                    if Subscription.objects.filter(profile=customer_profile, gateway_id=stripe_customer_subscription.id).exists():
                        stripe.sync_stripe_subscription(site, stripe_customer_subscription)
                        raise Exception(f"Stripe Subscription Synced {stripe_customer_subscription.id}")
                
                stripe_product = stripe.get_stripe_product(site, offer.pk)
                if not stripe_product:
                    stripe.create_offers([offer])

                if stripe_product and 'stripe' not in offer.meta:
                    offer.meta.update({'stripe': {'product_id': stripe_product.id}})
                    offer.save()
                
                stripe.sync_offer_prices(offer)
                
                stripe_payment_methods = stripe.get_customer_payment_methods(stripe_customer.id)
                if not stripe_payment_methods:
                    raise ObjectDoesNotExist(f"stripe_payment_method does not exist for stripe_customer: {stripe_customer.id}")
                
                setup_intent_object = stripe.build_setup_intent(stripe_payment_methods[0].id)
                stripe_setup_intent = stripe.stripe_create_object(stripe.stripe.SetupIntent, setup_intent_object)
                if not stripe_setup_intent:
                    raise Exception(f"Could not create stripe_setup_intent transaction_info: {stripe.transaction_info}")
                
                subscription_obj = stripe.build_subscription(stripe.invoice.order_items.first(), stripe_payment_methods[0].id)
                trial_days = offer.get_offer_end_date(last_transaction_datetime) - timezone.now()
                subscription_obj['trial_period_days'] = trial_days.days
                subscription_obj['billing_cycle_anchor'] = offer.get_offer_end_date(last_transaction_datetime)
                stripe_subscription = stripe.stripe_create_object(stripe.stripe.Subscription, subscription_obj)

                if not stripe_subscription or stripe_subscription.status == 'incomplete':
                    stripe.transaction_succeeded = False
                    logger.error(f"AuthorizeNet Subscription {auth_subscription.id.text} was not transfered to stripe error {stripe.transaction_info}")
                else:
                    subscription = Subscription.objects.create(
                        gateway_id=stripe_subscription.id,
                        profile=customer_profile,
                        auto_renew=True,
                        status=SubscriptionStatus.ACTIVE
                    )
                    subscription.meta['response'] = stripe.transaction_info
                    subscription.save()

            except ObjectDoesNotExist as exce:
                logger.exception(f"sync_subscriptions exception: {exce} subscription: ({auth_subscription.id.text}, {subscription_info.subscription.name})")
            except MultipleObjectsReturned as exce:
                logger.exception(f"sync_subscriptions exception: {exce} subscription: ({auth_subscription.id.text}, {subscription_info.subscription.name})")
            except Exception as exce:
                logger.exception(f"sync_subscriptions exception: {exce} subscription: ({auth_subscription.id.text}, {subscription_info.subscription.name})")



from core.models import Product
from datetime import timedelta

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.sites.models import Site
from django.utils import timezone
from django.urls import reverse
from django.test import TestCase, Client, tag

from unittest import skipIf
from random import randrange, choice
from siteconfigs.models import SiteConfigModel
from vendor.forms import CreditCardForm, BillingAddressForm
from vendor.models import Invoice, Payment, Offer, Price, Receipt, CustomerProfile, OrderItem, Subscription
from vendor.models.choice import PurchaseStatus, InvoiceStatus, SubscriptionStatus
from vendor.processors import PaymentProcessorBase, AuthorizeNetProcessor, StripeProcessor
###############################
# Test constants
###############################

User = get_user_model()


class BaseProcessorTests(TestCase):

    fixtures = ['user', 'unit_test']

    def setUp(self):
        self.client = Client()
        self.site = Site.objects.get(pk=1)
        self.user = User.objects.get(pk=1)
        self.client.force_login(self.user)
        self.existing_invoice = Invoice.objects.get(pk=1)
        self.base_processor = PaymentProcessorBase(self.site, self.existing_invoice)
        self.subscription_offer = Offer.objects.get(pk=4)
        self.hamster_wheel = Offer.objects.get(pk=3)
        self.form_data = {
            'billing_address_form': {
                'billing-name': 'Home',
                'billing-company': 'Whitemoon Dreams',
                'billing-country': '840',
                'billing-address_1': '221B Baker Street',
                'billing-address_2': '',
                'billing-locality': 'Marylebone',
                'billing-state': 'California',
                'billing-postal_code': '90292'
            },
            'credit_card_form': {
                'full_name': 'Bob Ross',
                'card_number': '5424000000000015',
                'expire_month': '12',
                'expire_year': '2030',
                'cvv_number': '900',
                'payment_type': '10'
            }
        }

    def test_base_processor_init_fail(self):
        with self.assertRaises(TypeError):
            base_processor = PaymentProcessorBase()

    def test_base_processor_init_success(self):
        base_processor = PaymentProcessorBase(self.site, self.existing_invoice)

        self.assertEquals('PaymentProcessorBase', base_processor.provider)
        self.assertIsNotNone(base_processor.invoice)

    def test_create_payment_model_success(self):
        self.base_processor.set_billing_address_form_data(self.form_data['billing_address_form'], BillingAddressForm)
        self.base_processor.set_payment_info_form_data(self.form_data['credit_card_form'], CreditCardForm)
        self.base_processor.is_data_valid()
        self.base_processor.create_payment_model()

        self.assertIsNotNone(self.base_processor.payment)

    def test_save_payment_transaction_success(self):
        payment_success = True
        transaction_id = '1423wasd'

        self.base_processor.set_billing_address_form_data(self.form_data['billing_address_form'], BillingAddressForm)
        self.base_processor.set_payment_info_form_data(self.form_data['credit_card_form'], CreditCardForm)
        self.base_processor.is_data_valid()
        self.base_processor.create_payment_model()
        self.base_processor.transaction_succeeded = payment_success
        self.base_processor.transaction_id = transaction_id
        self.base_processor.save_payment_transaction_result()

        self.assertIsNotNone(self.base_processor.payment)
        self.base_processor.payment.refresh_from_db()

        self.assertTrue(self.base_processor.payment.success)
        self.assertEquals(self.base_processor.payment.transaction, transaction_id)
        self.assertIn('payment_info', self.base_processor.payment.result)

    def test_update_invoice_status_success(self):
        self.base_processor.transaction_succeeded = True
        self.base_processor.update_invoice_status(InvoiceStatus.COMPLETE)

        self.assertEquals(InvoiceStatus.COMPLETE, self.base_processor.invoice.status)

    def test_update_invoice_status_fails(self):
        self.base_processor.update_invoice_status(InvoiceStatus.COMPLETE)

        self.assertNotEquals(InvoiceStatus.COMPLETE, self.base_processor.invoice.status)

    def test_create_receipt_by_term_type_subscription(self):
        self.base_processor.invoice.add_offer(self.subscription_offer)
        self.base_processor.invoice.save()

        order_item_subscription = self.base_processor.invoice.order_items.get(offer__pk=4)
        self.base_processor.payment = Payment.objects.get(pk=1)
        
        self.base_processor.subscription_id = "123"
        self.base_processor.create_subscription_model()
        self.base_processor.create_receipt_by_term_type(order_item_subscription, order_item_subscription.offer.terms)

        self.assertIsNotNone(self.base_processor.subscription)
        self.assertIsNotNone(self.base_processor.receipt.subscription)

    def test_create_receipt_by_term_type_perpetual(self):
        self.base_processor.invoice.save()
        perpetual_order_item = self.base_processor.invoice.order_items.get(offer__pk=1)

        self.base_processor.payment = Payment.objects.get(pk=1)
        self.base_processor.create_receipt_by_term_type(perpetual_order_item, perpetual_order_item.offer.terms)

        self.assertIsNone(self.base_processor.receipt.subscription)

    # def test_create_receipt_by_term_type_one_time_use(self):
        # raise NotImplementedError()

    def test_create_receipts_success(self):
        self.base_processor.invoice.status = InvoiceStatus.COMPLETE
        self.base_processor.payment = Payment.objects.get(pk=1)
        self.base_processor.create_receipts(self.base_processor.invoice.order_items.all())

        self.assertEquals(Receipt.objects.all().count(), sum([ order_item.receipts.all().count() for order_item in self.base_processor.invoice.order_items.all() ]))

    # def test_update_subscription_receipt_success(self):
    #     subscription_id = 123456789
    #     self.base_processor.invoice.add_offer(self.subscription_offer)
    #     self.base_processor.invoice.save()
    #     self.base_processor.invoice.status = InvoiceStatus.COMPLETE
    #     self.base_processor.payment = Payment.objects.get(pk=1)
    #     self.base_processor.create_receipts(self.base_processor.invoice.order_items.all())

    #     subscription_list = self.existing_invoice.order_items.filter(offer__terms=TermType.SUBSCRIPTION)
    #     subscription = subscription_list[0]

    #     self.base_processor.update_subscription_receipt(subscription, subscription_id, PurchaseStatus.SETTLED)
    #     receipt = Receipt.objects.get(meta__subscription_id=subscription_id)

    #     self.assertIsNotNone(receipt)
    #     self.assertEquals(subscription_id, receipt.meta['subscription_id'])

    def test_amount_success(self):
        self.existing_invoice.update_totals()
        self.assertEquals(self.existing_invoice.total, self.base_processor.amount())

    def test_amount_without_subscriptions_success(self):
        self.base_processor.invoice.add_offer(self.subscription_offer)

        price = Price()
        price.offer = self.subscription_offer
        price.cost = 25
        price.start_date = timezone.now() - timedelta(days=1)
        price.save()
        self.assertNotEquals(self.existing_invoice.total, self.base_processor.amount_without_subscriptions())

    def test_get_transaction_id_success(self):
        self.base_processor.payment = Payment.objects.get(pk=1)
        self.assertIn(str(settings.SITE_ID), self.base_processor.get_transaction_id())
        self.assertIn(str(self.existing_invoice.profile.pk), self.base_processor.get_transaction_id())
        self.assertIn(str(self.existing_invoice.pk), self.base_processor.get_transaction_id())

    def test_set_billing_address_form_data_fail(self):
        with self.assertRaises(TypeError):
            self.base_processor.set_billing_address_form_data(self.form_data)

    def test_set_billing_address_form_data_success(self):
        self.base_processor.set_billing_address_form_data(self.form_data['billing_address_form'], BillingAddressForm)

        self.assertIsNotNone(self.base_processor.billing_address)
        self.assertIn(self.form_data['billing_address_form']['billing-address_1'], self.base_processor.billing_address.data['billing-address_1'])

    def test_set_payment_info_form_data_fail(self):
        with self.assertRaises(TypeError):
            self.base_processor.set_payment_info_form_data(self.form_data)

    def test_set_payment_info_form_data_success(self):
        self.base_processor.set_payment_info_form_data(self.form_data['credit_card_form'], CreditCardForm)

        self.assertIsNotNone(self.base_processor.payment_info)
        self.assertIn(self.form_data['credit_card_form']['cvv_number'], self.base_processor.payment_info.data['cvv_number'])

    def test_get_checkout_context_success(self):
        context = self.base_processor.get_checkout_context()
        self.assertIn('invoice', context)

    def test_free_payment_success(self):
        customer = CustomerProfile.objects.get(pk=2)
        invoice = Invoice(profile=customer)
        invoice.save()
        invoice.add_offer(Offer.objects.get(pk=8))

        base_processor = PaymentProcessorBase(invoice.site, invoice)

        base_processor.set_billing_address_form_data(self.form_data['billing_address_form'], BillingAddressForm)
        base_processor.set_payment_info_form_data(self.form_data['credit_card_form'], CreditCardForm)

        base_processor.authorize_payment()

        self.assertTrue(invoice.payments.count())
        self.assertTrue(customer.receipts.count())

    def test_renew_subscription(self):
        subscription = Subscription.objects.get(pk=1)
        offer = subscription.get_offer()
        offer.term_details['trial_occurrences'] = 0
        offer.term_details['trial_days'] = 0
        offer.save()
        offer.refresh_from_db()
        subscription.save()
        subscription.refresh_from_db()

        submitted_datetime = timezone.now()

        invoice = Invoice.objects.create(
            profile=subscription.profile,
            site=subscription.profile.site,
            ordered_date=submitted_datetime,
            status=InvoiceStatus.COMPLETE
        )
        invoice.add_offer(subscription.receipts.first().order_item.offer)
        invoice.save()

        transaction_id = timezone.now().strftime("%Y-%m-%d_%H-%M-%S-Manual-Renewal")

        base_processor = PaymentProcessorBase(invoice.site, invoice)
        base_processor.renew_subscription(subscription, transaction_id, PurchaseStatus.CAPTURED)

        self.assertTrue(subscription.profile.has_product(subscription.receipts.last().products.all()))

    def test_subscription_price_update_success(self):
        subscription = Subscription.objects.get(pk=1)
        offer = Offer.objects.get(pk=4)
        price = Price.objects.create(offer=offer, cost=89.99, currency='usd', start_date=timezone.now())
        offer.prices.add(price)

        processor = PaymentProcessorBase(subscription.profile.site)
        processor.subscription_update_price(subscription, price, self.user)

        subscription.refresh_from_db()
        self.assertIn('price_update', subscription.meta)

    def test_subscription_payment_billing_start_date(self):
        today = timezone.now()
        self.subscription_offer.billing_start_date = today + timedelta(days=10)
        self.subscription_offer.save()

        self.base_processor.set_billing_address_form_data(self.form_data['billing_address_form'], BillingAddressForm)
        self.base_processor.set_payment_info_form_data(self.form_data['credit_card_form'], CreditCardForm)
        self.base_processor.is_data_valid()

        self.base_processor.invoice.add_offer(self.subscription_offer)
        self.base_processor.invoice.save()
        self.base_processor.transaction_succeeded = True
        self.base_processor.process_subscriptions()
    
        self.assertEqual(today.strftime("%y/%m/%d"), self.base_processor.trial_receipt.start_date.strftime("%y/%m/%d"))
        self.assertEqual((self.subscription_offer.billing_start_date - timedelta(days=1)).strftime("%y/%m/%d"), self.base_processor.trial_receipt.end_date.strftime("%y/%m/%d"))

    def test_subscription_payment_term_start_date(self):
        today = timezone.now()
        self.subscription_offer.term_start_date = today + timedelta(days=10)
        self.subscription_offer.term_details['trial_amount'] = 0
        self.subscription_offer.term_details['trial_occurrences'] = 0
        self.subscription_offer.term_details['trial_days'] = 0
        self.subscription_offer.save()

        self.base_processor.set_billing_address_form_data(self.form_data['billing_address_form'], BillingAddressForm)
        self.base_processor.set_payment_info_form_data(self.form_data['credit_card_form'], CreditCardForm)
        self.base_processor.is_data_valid()

        self.base_processor.invoice.add_offer(self.subscription_offer)
        self.base_processor.invoice.save()
        self.base_processor.transaction_succeeded = True
        self.base_processor.process_subscriptions()
    
        self.assertIsNone(self.base_processor.trial_receipt)
        self.assertEqual(self.subscription_offer.term_start_date, self.base_processor.receipt.start_date)
    
    def test_subscription_payment_trial_days(self):
        today = timezone.now()
        self.subscription_offer.term_start_date = today + timedelta(days=10)
        self.subscription_offer.term_details['trial_amount'] = 10
        self.subscription_offer.term_details['trial_occurrences'] = 0
        self.subscription_offer.term_details['trial_days'] = 7
        self.subscription_offer.save()

        self.base_processor.set_billing_address_form_data(self.form_data['billing_address_form'], BillingAddressForm)
        self.base_processor.set_payment_info_form_data(self.form_data['credit_card_form'], CreditCardForm)
        self.base_processor.is_data_valid()

        self.base_processor.invoice.add_offer(self.subscription_offer)
        self.base_processor.invoice.save()
        self.base_processor.transaction_succeeded = True
        self.base_processor.process_subscriptions()
    
        self.assertEqual(self.subscription_offer.term_start_date.strftime("%y/%m/%d"), self.base_processor.trial_receipt.start_date.strftime("%y/%m/%d"))
        self.assertEqual((self.base_processor.trial_receipt.end_date).strftime("%y/%m/%d"), (self.subscription_offer.term_start_date + timezone.timedelta(days=self.subscription_offer.term_details['trial_days'])).strftime("%y/%m/%d"))
    
    # def test_get_header_javascript_success(self):
    #     raise NotImplementedError()

    # def test_get_javascript_success(self):
    #     raise NotImplementedError()

    # def test_get_template_success(self):
    #     raise NotImplementedError()

    # def test_authorize_payment_success(self):
    #     raise NotImplementedError()

    # def test_pre_authorization_success(self):
    #     raise NotImplementedError()

    # def test_process_payment_success(self):
    #     raise NotImplementedError()

    # def test_post_authorization_success(self):
    #     raise NotImplementedError()

    # def test_capture_payment_success(self):
    #     raise NotImplementedError()

    # def test_subscription_payment_success(self):
    #     raise NotImplementedError()

    # def test_subscription_cancel_success(self):
    #     raise NotImplementedError()

    # def test_refund_payment_success(self):
    #     raise NotImplementedError()


class SupportedProcessorsSetupTests(TestCase):

    fixtures = ['user', 'unit_test']

    def setUp(self):
        self.invoice = Invoice.objects.get(pk=1)
        self.site = Site.objects.get(pk=1)

    def test_configured_processor_setup(self):
        """
        Test the initialized of the PaymentProcessor defined in the setting file
        """
        try:
            processor = PaymentProcessorBase(self.site, self.invoice)
        except Exception:
            print("Warning PaymentProcessor defined in settings file did not pass init")
        finally:
            pass

    def test_authorize_net_init(self):
        try:
            if not (settings.AUTHORIZE_NET_TRANSACTION_KEY and settings.AUTHORIZE_NET_API_ID):
                raise ValueError(
                "Missing Authorize.net keys in settings: AUTHORIZE_NET_TRANSACTION_KEY and/or AUTHORIZE_NET_API_ID")
            processor = AuthorizeNetProcessor(self.site, self.invoice)
        except Exception:
            print("AuthorizeNetProcessor did not initalized correctly")
        finally:
            pass

    # def test_stripe_init(self):
        # raise NotImplementedError()




from django.apps import apps
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import ObjectDoesNotExist
from django.forms import BaseModelForm
from django.http import HttpResponse, JsonResponse
from django.shortcuts import redirect, get_object_or_404
from django.urls import reverse_lazy, reverse
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.generic.edit import BaseUpdateView

from vendor.config import VENDOR_PRODUCT_MODEL
from vendor.forms import PaymentRefundForm
from vendor.models import CustomerProfile, Payment, Offer, Receipt, Subscription
from vendor.models.choice import InvoiceStatus
from vendor.processors import get_site_payment_processor
from vendor.utils import get_or_create_session_cart, get_site_from_request


Product = apps.get_model(VENDOR_PRODUCT_MODEL)


class VendorIndexAPI(View):
    """
    docstring
    """
    def get(self, request, *args, **kwargs):
        return HttpResponse('<h1>Welcome to vendor APIs<h1>')


class AddToCartView(View):
    '''
    Create an order item and adds it to the order. No login required as the request can come
    from a user without a session.
    '''
    def session_cart(self, request, offer):
        offer_key = str(offer.pk)
        session_cart = get_or_create_session_cart(request.session)

        if offer_key not in session_cart:
            session_cart[offer_key] = {}
            session_cart[offer_key]['quantity'] = 0

        session_cart[offer_key]['quantity'] += 1

        if not offer.allow_multiple:
            session_cart[offer_key]['quantity'] = 1

        return session_cart
    
    def add_offer_to_customer_profile_cart(self, customer_profile, offer):
        cart = customer_profile.get_cart_or_checkout_cart()
        all_products = offer.products.all()

        if cart.status == InvoiceStatus.CHECKOUT:
            cart.status = InvoiceStatus.CART
            cart.save()

        if customer_profile.has_product(all_products) and not offer.allow_multiple:
            messages.info(self.request, _("You Have Already Purchased This Item"))
        elif cart.order_items.filter(offer__products__in=all_products).exists() and not offer.allow_multiple:
            messages.info(self.request, _("You already have this product in you cart. You can only buy one"))
        else:
            messages.info(self.request, _("Added item to cart."))
            cart.add_offer(offer)

    def post(self, request, *args, **kwargs):
        try:
            offer = Offer.objects.get(site=get_site_from_request(request), slug=self.kwargs["slug"], available=True)
        except ObjectDoesNotExist:
            messages.error(request, _("Offer does not exist or is unavailable"))
            return redirect('vendor:cart')

        if not request.user.is_authenticated:
            request.session['session_cart'] = self.session_cart(request, offer)
            return redirect('vendor:cart')      # Redirect to cart on success

        profile, created = self.request.user.customer_profile.get_or_create(site=get_site_from_request(request))

        self.add_offer_to_customer_profile_cart(profile, offer)

        return redirect('vendor:cart')      # Redirect to cart on success
    
    def get(self, request, *args, **kwargs):
        try:
            offer = Offer.objects.get(site=get_site_from_request(request), slug=self.kwargs["slug"], available=True)
        except ObjectDoesNotExist:
            messages.error(_("Offer does not exist or is unavailable"))
            return redirect('vendor:cart')

        if not request.user.is_authenticated:
            request.session['session_cart'] = self.session_cart(request, offer)
            return redirect('vendor:cart')      # Redirect to cart on success

        profile, created = self.request.user.customer_profile.get_or_create(site=get_site_from_request(request))

        self.add_offer_to_customer_profile_cart(profile, offer)

        return redirect('vendor:cart')      # Redirect to cart on success


class RemoveFromCartView(View):
    '''
    Removes an order item and adds it to the order. No login required as the request can come
    from a user without a session.
    '''
    def post(self, request, *args, **kwargs):
        offer = get_object_or_404(Offer, site=get_site_from_request(request), slug=self.kwargs["slug"])
        if not request.user.is_authenticated:
            offer_key = str(offer.pk)
            session_cart = get_or_create_session_cart(request.session)

            if offer_key in session_cart:
                session_cart[offer_key]['quantity'] -= 1

            if session_cart[offer_key]['quantity'] <= 0:
                del(session_cart[offer_key])

            request.session['session_cart'] = session_cart
        else:
            profile = self.request.user.customer_profile.get(site=get_site_from_request(request))      # Make sure they have a cart

            cart = profile.get_cart_or_checkout_cart()

            if cart.status == InvoiceStatus.CHECKOUT:
                cart.status = InvoiceStatus.CART
                cart.save()

            cart.remove_offer(offer)

        messages.info(self.request, _("Removed item from cart."))
        return redirect('vendor:cart')      # Redirect to cart on success


class PaymentGatewaySubscriptionCancelView(LoginRequiredMixin, View):
    success_url = reverse_lazy('vendor:customer-subscriptions')

    def post(self, request, *args, **kwargs):
        subscription = get_object_or_404(Subscription, uuid=self.kwargs["uuid"])

        processor = get_site_payment_processor(subscription.profile.site)(subscription.profile.site)
        
        try:
            processor.subscription_cancel(subscription)
            messages.info(self.request, _("Subscription Cancelled"))

        except Exception:
            messages.warning(self.request, _("Cancel Subscription Failed"))

        return redirect(request.META.get('HTTP_REFERER', self.success_url))


class CustomerSubscriptionsCancelModel(LoginRequiredMixin, View):
    success_url = reverse_lazy('vendor:customer-subscriptions')

    def post(self, request, **kwargs):
        site = get_site_from_request(request)
        customer_profile = CustomerProfile.objects.get(site=site, user=request.user)

        for subscription in customer_profile.get_active_subscriptions():
            subscription.cancel()

        messages.info(self.request, _("Subscriptions Cancelled"))

        return redirect(request.META.get('HTTP_REFERER', self.success_url))


class VoidProductView(LoginRequiredMixin, View):
    success_url = reverse_lazy('vendor_admin:manage-profiles')

    def post(self, request, *args, **kwargs):
        receipt = get_object_or_404(Receipt, uuid=self.kwargs["uuid"])
        receipt.subscription.void()

        messages.info(request, _("Customer has no longer access to Product"))
        return redirect(request.META.get('HTTP_REFERER', self.success_url))


class AddOfferToProfileView(LoginRequiredMixin, View):

    def get(self, request, *args, **kwargs):
        customer_profile = get_object_or_404(CustomerProfile, uuid=kwargs.get('uuid_profile'))
        offer = get_object_or_404(Offer, uuid=kwargs['uuid_offer'])

        cart = customer_profile.get_cart_or_checkout_cart()
        cart.empty_cart()
        cart.add_offer(offer)

        if offer.current_price() or cart.total:
            messages.info(request, _("Offer and Invoice must have zero value"))
            cart.remove_offer(offer)
            return redirect(reverse('vendor_admin:manager-profile', kwargs={'uuid': customer_profile.uuid}))

        processor = get_site_payment_processor(cart.site)(cart.site, cart)
        processor.authorize_payment()

        messages.info(request, _("Offer Added To Customer Profile"))
        return redirect(reverse('vendor_admin:manager-profile', kwargs={'uuid': customer_profile.uuid}))


class ProductAvailabilityToggleView(LoginRequiredMixin, View):
    success_url = reverse_lazy('vendor_admin:manager-product-list')

    def post(self, request, *args, **kwargs):
        product = get_object_or_404(Product, uuid=self.kwargs.get('uuid'))

        product.available = request.POST.get('available', False)
        product.save()

        for offer in Offer.objects.filter(products__in=[product]):
            offer.available = product.available
            offer.save()

        messages.info(request, _("Product availability Changed"))
        return redirect(request.META.get('HTTP_REFERER', self.success_url))


class SubscriptionPriceUpdateView(LoginRequiredMixin, View):
    success_url = reverse_lazy('vendor_admin:manager-product-list')

    def post(self, request, *args, **kwargs):
        site = get_site_from_request(request)
        subscription = get_object_or_404(Subscription, profile__site=site, uuid=self.request.POST.get('subscription_uuid'))
        offer = get_object_or_404(Offer, site=site, uuid=self.request.POST.get('offer_uuid'))

        processor = get_site_payment_processor(site)(site)
        processor.subscription_update_price(subscription, offer.current_price(), request.user)
        
        return redirect(request.META.get('HTTP_REFERER', self.success_url))


class RenewSubscription(LoginRequiredMixin, View):

    def post(self, request, *args, **kwargs):
        # TODO: Need to implement
        pass


class RefundPaymentAPIView(LoginRequiredMixin, BaseUpdateView):
    form_class = PaymentRefundForm
    success_url = reverse_lazy("vendor:customer-subscriptions")

    def get_object(self):
        return Payment.objects.get(
            uuid=self.kwargs.get("uuid"),
            invoice__site=get_site_from_request(self.request),
        )
    
    def get(self, request, *args, **kwargs):
        payment = self.get_object()
        refund_form = self.get_form_class()(instance=payment)

        return JsonResponse(refund_form.data)

    def form_valid(self, form):
        processor = get_site_payment_processor(form.instance.invoice.site)(
            form.instance.invoice.site
        )

        try:
            processor.refund_payment(form)
            if not processor.transaction_succeeded:
                return JsonResponse({"error": processor.transaction_info})

        except Exception as exc:
            return JsonResponse({"error": str(exc)})

        return JsonResponse({"message": _("Payment Refunded")})

    def form_invalid(self, form: BaseModelForm):
        return JsonResponse({"error": form.errors})



"""
Base Payment processor used by all derived processors.
"""
import logging
from datetime import timedelta
from decimal import Decimal

import django.dispatch
from django.utils import timezone

from vendor import config
from vendor.forms import BillingAddressForm, CreditCardForm
from vendor.models import Payment, Receipt, Subscription
from vendor.models.choice import (InvoiceStatus, PaymentTypes, PurchaseStatus,
                                  SubscriptionStatus, TermType)

logger = logging.getLogger(__name__)


##########
# SIGNALS
vendor_pre_authorization = django.dispatch.Signal()
vendor_process_payment = django.dispatch.Signal()
vendor_post_authorization = django.dispatch.Signal()
vendor_subscription_cancel = django.dispatch.Signal()
vendor_customer_card_expiring = django.dispatch.Signal()


#############
# BASE CLASS
class PaymentProcessorBase(object):
    """
    Setup the core functionality for all processors.
    """
    API_ENDPOINT = None

    status = None
    invoice = None
    provider = None
    # TODO: Change payment to a list as an invoice can have multiple payment. EG: 1 for 1 type purchases and n for any amount of subscriptions. 
    payment = None
    subscription = None
    subscription_id = None
    receipt = None
    trial_receipt = None
    payment_info = {}
    billing_address = {}
    transaction_token = None
    transaction_id = ""
    transaction_succeeded = False
    transaction_info = {}
    transaction_response = None

    def __init__(self, site, invoice=None):
        """
        This should not be overriden.  Override one of the methods it calls if you need to.
        """
        if invoice:
            self.set_invoice(invoice)

        self.provider = self.__class__.__name__
        self.processor_setup(site)
        self.set_api_endpoint()

    def set_api_endpoint(self):
        """
        Sets the API endpoint for debugging or production.It is dependent on the VENDOR_STATE
        enviornment variable. Default value is DEBUG for the VENDOR_STATE this function
        should be overwrote upon necesity of each Payment Processor
        """
        if config.VENDOR_STATE == 'DEBUG':
            self.API_ENDPOINT = None
        elif config.VENDOR_STATE == 'PRODUCTION':
            self.API_ENDPOINT = None

    def processor_setup(self, site):
        """
        This is for setting up any of the settings needed for the payment processing.
        For example, here you would set the
        """
        pass

    def set_payment_info(self, **kwargs):
        self.payment_info = kwargs

    def set_invoice(self, invoice):
        self.invoice = invoice

    def create_payment_model(self, amount=None):
        """
        Create payment instance with base information to track payment submissions
        """
        if not amount:
            amount = self.invoice.total

        self.payment = Payment(profile=self.invoice.profile,
                               amount=amount,
                               provider=self.provider,
                               invoice=self.invoice,
                               created=timezone.now()
                               )

        if not self.payment.result:
            self.payment.result = {}
        
        self.payment.result['payment_info'] = {
            'account_number': self.payment_info.cleaned_data.get('card_number')[-4:],
            'full_name': self.payment_info.cleaned_data.get('full_name')
        }
        self.payment.payee_full_name = self.payment_info.cleaned_data.get('full_name')
        self.payment.payee_company = self.billing_address.cleaned_data.get('company')
        self.payment.status = PurchaseStatus.QUEUED
        self.payment.submitted_date = timezone.now()

        billing_address = self.billing_address.save(commit=False)
        billing_address, created = self.invoice.profile.get_or_create_address(billing_address)

        if created:
            billing_address.profile = self.invoice.profile
            billing_address.save()

        self.payment.billing_address = billing_address
        self.payment.save()

    def save_payment_transaction_result(self):
        """
        Saves the result output of any transaction.
        """
        self.payment.success = self.transaction_succeeded
        self.payment.transaction = self.transaction_id
        self.payment.result.update(self.transaction_info)
        self.payment.save()

    def get_payment_info(self, account_number=None, full_name=None):
        """
        Each processor should implement their own method, but they should
        all return at least the account_number and full_name as a dictionary.
        eg:
        """
        return {
            'account_number': account_number,
            'full_name': full_name
        }

    def get_transaction_info(self, raw='', errors='', payment_method='', data=''):
        return {
            'raw': raw,
            'errors': errors,
            'payment_method': payment_method,
            'data': data
        }

    def parse_response(self):
        ...

    def parse_success(self):
        ...

    def save_subscription_transaction_result(self):
        """
        Saves the result output of any transaction.
        """
        if self.transaction_succeeded:
            self.subscription.status = SubscriptionStatus.ACTIVE
            self.subscription.gateway_id = self.transaction
            
        self.subscription.meta[timezone.now().strftime("%Y-%m-%d_%H:%M:%S")] = self.transaction_info
        self.subscription.save()

    def update_invoice_status(self, new_status):
        """
        Updates the Invoice status if the transaction was submitted.
        Otherwise it returns the invoice to the Cart. The error is saved in
        the payment for the transaction.
        """
        if self.transaction_succeeded:
            self.invoice.status = new_status
        else:
            self.invoice.status = InvoiceStatus.CART

        self.invoice.save()

    def is_transaction_and_invoice_complete(self):
        """
        If payment was successful and invoice status is complete returns True. Otherwise
        false and no receipts should be created.
        """
        if self.transaction_succeeded and self.invoice.status == InvoiceStatus.COMPLETE:
            return True

        return False

    def create_receipt_by_term_type(self, order_item, term_type):
        today = timezone.now()
        self.receipt = Receipt()
        self.receipt.profile = self.invoice.profile
        self.receipt.order_item = order_item
        self.receipt.transaction = self.payment.transaction
        self.receipt.meta.update(self.payment.result)
        self.receipt.meta['payment_amount'] = str(self.to_valid_decimal(self.payment.amount))

        if today > (self.payment.submitted_date - timedelta(hours=1)) or today < (self.payment.submitted_date + timedelta(hours=1)):
            start_date = order_item.offer.get_offer_start_date(self.payment.submitted_date)
        else:
            start_date = order_item.offer.get_offer_start_date(today)


        if self.trial_receipt:
            start_date = self.trial_receipt.end_date

        self.receipt.start_date = start_date
        self.receipt.save()

        if term_type < TermType.PERPETUAL:
            self.receipt.end_date = order_item.offer.get_offer_end_date(self.receipt.start_date)
            self.receipt.subscription = self.subscription
            
        self.receipt.save()

    def create_trial_receipt_payment(self, order_item):
        today = timezone.now()

        if today > (self.payment.submitted_date - timedelta(hours=1)) or today < (self.payment.submitted_date + timedelta(hours=1)):
            start_date = order_item.offer.get_offer_start_date(self.payment.submitted_date)
        else:
            start_date = order_item.offer.get_offer_start_date(today)

        end_date = order_item.offer.get_trial_end_date(start_date)

        if end_date == start_date:
            self.trial_payment = None
            self.trial_receipt = None
            return None   # There is something wrong if both start and end date are the same

        self.trial_payment, created = Payment.objects.get_or_create(
            profile=self.invoice.profile,
            amount=order_item.offer.get_trial_amount(),
            provider=self.provider,
            invoice=self.invoice,
            defaults={
                "submitted_date": start_date,
                "payee_full_name": " ".join([self.invoice.profile.user.first_name, self.invoice.profile.user.last_name])
            }
        )
        
        self.trial_payment.transaction = f"{self.trial_payment.uuid}-trial"
        self.trial_payment.status = PurchaseStatus.SETTLED
        self.trial_payment.success = True
        self.trial_payment.save()

        self.trial_receipt = Receipt.objects.create(
            profile=self.invoice.profile,
            order_item=order_item,
            transaction=self.trial_payment.transaction,
            start_date=start_date,
            end_date=end_date,
            subscription=self.subscription
        )

        if order_item.offer.terms < TermType.PERPETUAL:
            self.trial_payment.subscription = self.subscription
            self.trial_payment.save()
            
            self.trial_receipt.subscription = self.subscription
            self.trial_receipt.save()
            
    def create_order_item_receipt(self, order_item):
        """
        Creates a receipt for every product in the order item according to its,
        offering term type.
        """
        for product in order_item.offer.products.all():
            if (order_item.offer.has_trial() or order_item.offer.has_valid_billing_start_date()) and\
                not self.invoice.profile.has_owned_product(order_item.offer.products.all()):
                self.create_trial_receipt_payment(order_item)
                if self.trial_receipt:
                    self.trial_receipt.products.add(product)
            else:
                self.create_receipt_by_term_type(order_item, order_item.offer.terms)
                self.receipt.products.add(product)

    def create_receipts(self, order_items):
        """
        It then creates receipt for the order items supplied.
        """
        for order_item in order_items.all():
            self.create_order_item_receipt(order_item)

    def update_subscription_receipt(self, subscription, subscription_id, status):
        """
        subscription: OrderItem
        subscription_id: int
        status: PurchaseStatus
        """
        subscription_receipt = self.invoice.order_items.get(offer=subscription.offer).receipts.get(transaction=self.payment.transaction)
        subscription_receipt.meta['subscription_id'] = subscription_id
        subscription_receipt.status = status
        subscription_receipt.save()

    def amount(self):   # Retrieves the total amount from the invoice
        self.invoice.update_totals()
        return self.invoice.total

    def amount_without_subscriptions(self):
        subscription_total = sum([ oi.total for oi in self.invoice.order_items.filter(offer__terms=TermType.SUBSCRIPTION)])

        amount = self.invoice.total - subscription_total
        return amount

    def get_transaction_id(self):
        return f"{self.invoice.site.pk}-{self.invoice.pk}"

    def set_billing_address_form_data(self, form_data, form_class):
        self.billing_address = form_class(form_data)

    def set_payment_info_form_data(self, form_data, form_class):
        self.payment_info = form_class(form_data)

    def is_data_valid(self):
        if not (self.billing_address.is_valid() and self.payment_info.is_valid() and self.invoice and self.invoice.order_items.count()):
            return False
        return True

    # -------------------
    # Data for the View
    def get_checkout_context(self, request=None, context={}):
        '''
        The Invoice plus any additional values to include in the payment record.
        '''
        # context = deepcopy(context)
        context['invoice'] = self.invoice
        
        if 'credit_card_form' not in context:
            context['credit_card_form'] = CreditCardForm(initial={'payment_type': PaymentTypes.CREDIT_CARD})

        if 'billing_address_form' not in context:
            context['billing_address_form'] = BillingAddressForm()

        return context

    def get_header_javascript(self):
        """
        Scripts that are expected to show in the top of the template.

        This will return a list of relative static URLs to the scripts.
        """
        return []

    def get_javascript(self):
        """
        Scripts added to the bottom of the page in the normal js location.

        This will return a list of relative static URLs to the scripts.
        """
        return []

    def get_template(self):
        """
        Unique partial template for the processor
        """
        pass

    def to_valid_decimal(self, number):
        # TODO: Need to check currency to determin decimal places.
        return Decimal(number).quantize(Decimal('.00'))

    def to_stripe_valid_unit(self, number):
        if number > 0:
            return int(number) * 100
        return 0

    # -------------------
    # Process a Payment
    def authorize_payment(self):
        """
        This runs the chain of events in a transaction.
        This should not be overriden.  Override one of the methods it calls if you need to.
        """
        self.invoice.ordered_date = timezone.now()
        self.invoice.save()
        if not self.invoice.calculate_subtotal():
            self.free_payment()
            return None

        if not self.is_data_valid():
            return None

        self.status = PurchaseStatus.QUEUED     # TODO: Set the status on the invoice.  Processor status should be the invoice's status.
        vendor_pre_authorization.send(sender=self.__class__, invoice=self.invoice)

        self.pre_authorization()

        self.status = PurchaseStatus.ACTIVE     # TODO: Set the status on the invoice.  Processor status should be the invoice's status.
        vendor_process_payment.send(sender=self.__class__, invoice=self.invoice)

        if self.invoice.get_one_time_transaction_order_items():
            self.create_payment_model()
            self.process_payment()
            self.save_payment_transaction_result()
            self.update_invoice_status(InvoiceStatus.COMPLETE)
            if self.is_transaction_and_invoice_complete():
                self.invoice.save_discounts_vendor_notes()
                self.create_receipts(self.invoice.get_one_time_transaction_order_items())

        if self.invoice.get_recurring_order_items():
            self.process_subscriptions()

        vendor_post_authorization.send(sender=self.__class__, invoice=self.invoice)
        self.post_authorization()

    def pre_authorization(self):
        """
        Called before the authorization begins.
        """
        pass

    def process_payment(self):
        """
        Called to handle the authorization.
        This is where the core of the payment processing happens.
        """
        # Gateway Transaction goes here...
        pass

    def free_payment(self):
        """
        Called to handle an invoice with total zero.
        This are the base internal steps to process a free payment.
        """
        self.payment = Payment(profile=self.invoice.profile,
                               amount=self.invoice.total,
                               provider=self.provider,
                               invoice=self.invoice,
                               created=timezone.now(),
                               submitted_date=timezone.now()
                               )
        self.payment.save()
        self.transaction_succeeded = True
        self.payment.success = True
        self.payment.status = PurchaseStatus.SETTLED
        self.payment.transaction = f"{self.payment.uuid}-free"
        self.payment.payee_full_name = " ".join([self.invoice.profile.user.first_name, self.invoice.profile.user.last_name])
        self.payment.result.update({'first': True})
        self.payment.save()

        self.update_invoice_status(InvoiceStatus.COMPLETE)

        self.create_receipts(self.invoice.order_items.all())

    def post_authorization(self):
        """
        Called after the authorization is complete.
        """
        pass

    def capture_payment(self):
        """
        Called to handle the capture.  (some gateways handle this at the same time as authorize_payment() )
        """
        pass

    def void_payment(self):
        """
        Call to handle a payment that has not been settled and wants to be voided
        """
        self.payment.status = PurchaseStatus.VOID
        self.payment.save()

    # -------------------
    # Process a Subscription
    def process_subscriptions(self):
        """
        Process/subscribies recurring payments throught the payement gateway and creates a payment model for each subscription.
        If a payment is completed it will create a receipt for the subscription
        """
        if not self.is_card_valid():
            return None

        for subscription in self.invoice.get_recurring_order_items():
            self.create_payment_model(amount=self.to_valid_decimal(subscription.total - subscription.discounts))
            self.subscription_payment(subscription)
            self.save_payment_transaction_result()
            self.update_invoice_status(InvoiceStatus.COMPLETE)
            
            if self.is_transaction_and_invoice_complete():
                self.create_subscription_model()
                self.invoice.save_discounts_vendor_notes()
                self.create_order_item_receipt(subscription)

    def subscription_payment(self, subscription):
        """
        Call handels the authrization and creation for a subscription.
        """
        # Gateway Transaction goes here...
        self.subscription_id = 'Test ID'
        ...

    def create_subscription_model(self):
        self.subscription = Subscription.objects.create(
            gateway_id=self.subscription_id,
            profile=self.invoice.profile,
            auto_renew=True,
            status=SubscriptionStatus.ACTIVE
        )
        self.subscription.meta['response'] = self.transaction_info
        self.subscription.save()
        self.payment.subscription = self.subscription
        self.payment.save()

    def subscription_info(self):
        pass

    def subscription_update_payment(self):
        pass

    def subscription_cancel(self, subscription):
        subscription.cancel()
        vendor_subscription_cancel.send(sender=self.__class__, subscription=subscription)

    def is_card_valid(self):
        """
        Function to validate a credit card by method of makeing a microtransaction and voiding it if authorized.
        """
        return True

    def renew_subscription(self, subscription, payment_transaction_id="", payment_status=PurchaseStatus.QUEUED, payment_success=True, submitted_date=timezone.now()):
        """
        Function to renew already paid subscriptions form the payment gateway provider.
        """
        if Payment.objects.filter(
                profile=subscription.profile,
                invoice=self.invoice,
                transaction=payment_transaction_id,
                subscription=subscription).exists():
            return None
        
        if Receipt.objects.filter(
            transaction=payment_transaction_id,
            profile=subscription.profile,
            subscription=subscription
        ).exists():
            return None

        self.subscription = subscription

        self.payment = Payment.objects.create(
            profile=subscription.profile,
            invoice=self.invoice,
            transaction=payment_transaction_id,
            subscription=subscription,
            submitted_date=submitted_date,
            amount=self.invoice.total,
            success=payment_success,
            status=payment_status,
            payee_full_name=" ".join([self.invoice.profile.user.first_name, self.invoice.profile.user.last_name])
        )
            
        if payment_status in [PurchaseStatus.QUEUED, PurchaseStatus.CAPTURED, PurchaseStatus.AUTHORIZED, PurchaseStatus.SETTLED]:
            self.create_receipts(self.invoice.order_items.all())

    def subscription_update_price(self, subscription, new_price, user):
        """
        Call to handle when a new subscription price needs to be approved.
        """
        now = timezone.now().strftime("%Y-%m-%d_%H:%M:%S")

        subscription.meta['price_update'] = {
            now: f'Price update ({new_price}) accepted by user: {user.username} on {now}'
        }

        subscription.save()
    
    # -------------------
    # Refund a Payment
    def refund_payment(self, refund_form, date=timezone.now()):
        refund_form.instance.record_refund(refund_form.cleaned_data['refund_amount'], date, refund_form.cleaned_data['reason'])
        
        if refund_form.cleaned_data['void_end_date']:
            receipt = refund_form.instance.get_receipt()
            receipt.end_date = date
            receipt.save()
        
        self.transaction_succeeded = True

    def subscription_payment_failed(self, subscription, transaction_id):
        self.payment = Payment.objects.create(
            subscription=subscription,
            profile=self.invoice.profile,
            amount=self.invoice.total,
            provider=self.provider,
            invoice=self.invoice,
            submitted_date=self.invoice.ordered_date,
            transaction=transaction_id,
            status=PurchaseStatus.DECLINED,
            payee_full_name=" ".join([self.invoice.profile.user.first_name, self.invoice.profile.user.last_name])
        )

    # -------------------
    # Charge a Customer Profile
    def charge_customer_profile(self):
        """
        Each processor need to implement this functions
        """
        ...
        
    def process_customer_profile_payment(self):
        """
        This runs the chain of events in a transaction.
        This should not be overriden.  Override one of the methods it calls if you need to.
        """
        self.invoice.ordered_date = timezone.now()
        self.invoice.save()
        if not self.invoice.calculate_subtotal():
            self.free_payment()
            return None

        if not self.is_data_valid():
            return None

        self.status = PurchaseStatus.QUEUED     # TODO: Set the status on the invoice.  Processor status should be the invoice's status.
        vendor_pre_authorization.send(sender=self.__class__, invoice=self.invoice)

        self.pre_authorization()

        self.status = PurchaseStatus.ACTIVE     # TODO: Set the status on the invoice.  Processor status should be the invoice's status.
        vendor_process_payment.send(sender=self.__class__, invoice=self.invoice)

        self.create_payment_model()
        self.charge_customer_profile()
        self.save_payment_transaction_result()
        self.update_invoice_status(InvoiceStatus.COMPLETE)
        if self.is_transaction_and_invoice_complete():
            self.invoice.save_discounts_vendor_notes()
            self.create_receipts(self.invoice.get_one_time_transaction_order_items())

        vendor_post_authorization.send(sender=self.__class__, invoice=self.invoice)
        self.post_authorization()

    ##########
    # Signals
    ##########
    def customer_card_expired(self, site, email):
        vendor_customer_card_expiring.send(sender=self.__class__, site_pk=site.pk, email=email)



import logging

from django.apps import apps
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import ObjectDoesNotExist
from django.db import IntegrityError, DatabaseError, transaction
from django.db.models import Count, Q
from django.shortcuts import redirect, render
from django.urls import reverse_lazy, reverse
from django.views.generic import TemplateView
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, UpdateView, FormMixin
from django.views.generic.list import ListView
from django.utils import timezone
from django.utils.translation import gettext as _

from vendor.config import VENDOR_PRODUCT_MODEL, SiteSelectForm

from vendor.forms import OfferForm, PriceFormSet, CreditCardForm, AddressForm, \
    SubscriptionForm, SiteSelectForm, SubscriptionAddPaymentForm, OfferSiteSelectForm, \
    StartDateForm

from vendor.models import Invoice, Offer, Receipt, CustomerProfile, Payment, Subscription
from vendor.models.choice import PaymentTypes, InvoiceStatus, PurchaseStatus
from vendor.views.mixin import PassRequestToFormKwargsMixin, SiteOnRequestFilterMixin, TableFilterMixin, get_site_from_request
from vendor.processors import get_site_payment_processor, StripeProcessor, PRORATION_BEHAVIOUR_CHOICE

Product = apps.get_model(VENDOR_PRODUCT_MODEL)
logger = logging.getLogger(__name__)


#############
# Admin Views
class AdminDashboardView(LoginRequiredMixin, SiteOnRequestFilterMixin, ListView):
    '''
    List of the most recent invoices generated on the current site.
    '''
    template_name = "vendor/manage/dashboard.html"
    model = Invoice

    def get_queryset(self):
        """
        Return the most recent 10
        """
        queryset = super().get_queryset()

        return queryset[:10]


class AdminInvoiceListView(LoginRequiredMixin, SiteOnRequestFilterMixin, ListView):
    '''
    List of all the invoices generated on the current site.
    '''
    template_name = "vendor/manage/invoice_list.html"
    model = Invoice

    def get_queryset(self):
        """
        Ignores Cart state invoices
        """
        queryset = super().get_queryset()
        return queryset.order_by('updated')


class AdminInvoiceDetailView(LoginRequiredMixin, DetailView):
    '''
    Details of an invoice generated on the current site.
    '''
    template_name = "vendor/manage/invoice_detail.html"
    model = Invoice
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'


class AdminProductListView(LoginRequiredMixin, TableFilterMixin, SiteOnRequestFilterMixin, ListView):
    '''
    Creates a Product to be added to offers
    '''
    template_name = "vendor/manage/products.html"
    model = Product
    paginate_by = 25

    def search_filter(self, queryset):
        return queryset.filter(name__icontains=self.request.GET.get('search_filter'))


class AdminProductUpdateView(LoginRequiredMixin, UpdateView):
    '''
    Details of an invoice generated on the current site.
    '''
    template_name = "vendor/manage/product.html"
    model = Product
    fields = ['sku', 'name', 'site', 'available', 'description', 'meta']
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'
    success_url = reverse_lazy('vendor_admin:manager-product-list')


class AdminProductCreateView(LoginRequiredMixin, CreateView):
    '''
    Creates a Product to be added to offers
    '''
    template_name = "vendor/manage/product.html"
    model = Product
    fields = ['sku', 'name', 'site', 'available', 'description', 'meta']
    success_url = reverse_lazy('vendor_admin:manager-product-list')

    def form_valid(self, form):
        new_product = form.save(commit=False)
        new_product.save()
        return redirect(self.success_url)


class AdminOfferListView(LoginRequiredMixin, SiteOnRequestFilterMixin, ListView):
    '''
    Creates a Product to be added to offers
    '''
    template_name = "vendor/manage/offers.html"
    model = Offer


class AdminOfferUpdateView(LoginRequiredMixin, PassRequestToFormKwargsMixin, UpdateView):
    '''
    Details of an invoice generated on the current site.
    '''
    template_name = "vendor/manage/offer.html"
    model = Offer
    form_class = OfferForm
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'
    template_name_suffix = '_update_form'

    def get_initial(self):
        return {'products': self.object.products.all()}

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        offer_products = self.object.products.all()
        customers_who_own = CustomerProfile.objects.filter(
            receipts__products__in=offer_products)
        customers_who_dont_own = CustomerProfile.objects.all().exclude(
            pk__in=[customer_profile.pk for customer_profile in customers_who_own.all()])

        context['customers_who_own'] = customers_who_own
        context['customers_who_dont_own'] = customers_who_dont_own

        context['formset'] = PriceFormSet(instance=self.object)

        return context

    def form_valid(self, form):
        price_formset = PriceFormSet(
            self.request.POST, self.request.FILES, instance=Offer.objects.get(uuid=self.kwargs['uuid']))

        offer = form.save(commit=False)

        if len(form.cleaned_data['products']) > 1:
            offer.bundle = True
        offer.save()

        for product in form.cleaned_data['products']:
            offer.products.add(product)

        if price_formset.has_changed() and not price_formset.is_valid():
            return render(self.request, self.template_name, {'form': form, 'formset': price_formset})
        elif price_formset.is_valid():
            for price_form in price_formset:
                price = price_form.save(commit=False)
                price.offer = offer
                if price_form.cleaned_data['price_select'] == 'free':
                    price.cost = 0
                price.save()
        else:
            return render(self.request, self.template_name, {'form': form, 'formset': price_formset})

        return redirect('vendor_admin:manager-offer-list')


class AdminOfferCreateView(LoginRequiredMixin, PassRequestToFormKwargsMixin, CreateView):
    '''
    Creates a Product to be added to offers
    '''
    template_name = "vendor/manage/offer.html"
    model = Offer
    form_class = OfferForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context['formset'] = PriceFormSet()

        return context

    def post(self, request):
        offer_form = self.form_class(request.POST, request=request)
        price_formset = PriceFormSet(request.POST)

        if not offer_form.is_valid():
            return render(request, self.template_name, {'form': offer_form, 'formset': price_formset})

        offer = offer_form.save(commit=False)
        if len(offer_form.cleaned_data['products']) > 1:
            offer.bundle = True

        offer.save()
        for product in offer_form.cleaned_data['products']:
            offer.products.add(product)

        if price_formset.has_changed() and not price_formset.is_valid():
            return render(request, self.template_name, {'form': offer_form, 'formset': price_formset})
        elif price_formset.has_changed() and price_formset.is_valid():
            product_currencies = {}
            # price_currency = [price_form.cleaned_data['currency']
            #                   for price_form in price_formset]

            for product in Product.objects.filter(pk__in=offer_form.cleaned_data['products']):
                for currency in product.meta['msrp'].keys():
                    product_currencies[currency] = currency

            for price_form in price_formset:
                if price_form.cleaned_data['currency'] not in product_currencies:
                    price_formset[0].add_error(
                        'currency', _('Invalid currency'))
                    return render(request, self.template_name, {'form': offer_form, 'formset': price_formset})

            for price_form in price_formset:
                price = price_form.save(commit=False)
                if price_form.cleaned_data['price_select'] == 'free':
                    price.cost = 0
                price.offer = offer
                price.save()

        return redirect('vendor_admin:manager-offer-list')


class AdminSubscriptionListView(LoginRequiredMixin, ListView):
    '''
    List of all the invoices generated on the current site.
    '''
    template_name = "vendor/manage/receipt_list.html"
    model = Subscription

    def get_queryset(self):
        queryset = super().get_queryset()
        return queryset.filter(profile__site=get_site_from_request(self.request))


class AdminStripeSubscriptionListView(LoginRequiredMixin, ListView):
    '''
    Lists Stripe Subscriptions from the requests, site.
    '''
    template_name = "vendor/manage/receipt_list.html"
    model = Subscription

    def get_queryset(self):
        site = get_site_from_request(self.request)
        qs = Subscription.objects.filter(profile__site=site, gateway_id__startswith="sub_").order_by("status", "-pk")
        return qs


class AdminStripeSubscriptionReCreate(LoginRequiredMixin, FormMixin, DetailView):
    '''
    View will get the Stripe Subscription Detail and on post
    will cancel the existing Subscription and create a new one
    to update any application fees or other price or offer
    related details
    '''
    template_name = 'vendor/manage/stripe_subscription_detail.html'
    model = Subscription
    slug_field = "uuid"
    slug_url_kwarg = "uuid"
    form_class = StartDateForm
    success_url = reverse_lazy('vendor_admin:manager-stripe-subscriptions')
    
    def post(self, request, **kwargs):
        form = StartDateForm(request.POST)

        if not form.is_valid():
            context = self.get_context_data(**kwargs)
            context['form'] = form
            return render(request, self.template_name, context)
        
        site = get_site_from_request(request)
        subscription = self.get_object()
        subscription_extras = {
            'billing_cycle_anchor': form.cleaned_data['start_date'],
            'proration_behavior': PRORATION_BEHAVIOUR_CHOICE.NONE.value
        }
        invoice = subscription.profile.get_cart()
        invoice.empty_cart()
        invoice.add_offer(subscription.get_offer())

        stripe = StripeProcessor(site, invoice)
        
        stripe_payment_methods = stripe.get_customer_payment_methods(subscription.profile.meta["stripe_id"])
        if not stripe_payment_methods:
            raise ObjectDoesNotExist(f"stripe_payment_method does not exist for stripe_customer: {subscription.profile.meta['stripe_id']}")
        
        stripe.create_subscription(stripe_payment_methods[0].id, subscription_extras)

        if stripe.transaction_succeeded:
            stripe.subscription_cancel(subscription)
        
        invoice.empty_cart()
        
        return redirect(reverse("vendor_admin:manager-subscription", kwargs={"uuid": str(self.get_object().uuid)}))


class AdminSubscriptionDetailView(LoginRequiredMixin, DetailView):
    '''
    Gets all Customer Profile information for quick lookup and management
    '''
    template_name = 'vendor/manage/subscription_detail.html'
    model = Subscription
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        subscription = context['object']
        payment = subscription.payments.first()

        context['payment'] = payment
        context['payment_form'] = CreditCardForm(
            initial={'payment_type': PaymentTypes.CREDIT_CARD}
        )
        if payment and payment.billing_address:
            context['billing_form'] = AddressForm(instance=payment.billing_address)

        context['payments'] = subscription.payments.order_by('-invoice__pk')
        context['receipts'] = subscription.receipts.order_by('-order_item__invoice__pk')

        return context


class AdminSubscriptionCreateView(LoginRequiredMixin, TemplateView):
    '''
    Gets all Customer Profile information for quick lookup and management
    '''
    template_name = 'vendor/manage/subscription_create.html'
    success_url = reverse_lazy('vendor_admin:manager-subscription-create')

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)

        if 'site' in request.GET:
            site_form = SiteSelectForm(request.GET)

            if not site_form.is_valid():
                context['site_form'] = site_form
                return render(request, self.template_name, context)
            
            context['subscription_form'] = SubscriptionForm(initial={'site': site_form.cleaned_data['site']})
            context['site_form'] = site_form

            return render(request, self.template_name, context)

        context['site_form'] = SiteSelectForm()

        return render(request, self.template_name, context)

    def post(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)

        subscription_form = SubscriptionForm(request.POST)
        
        if not subscription_form.is_valid():
            context['subscription_form'] = subscription_form
            return render(request, self.template_name, context)

        subscription = Subscription.objects.create(
            profile=subscription_form.cleaned_data['profile'],
            gateway_id=subscription_form.cleaned_data['subscription_id'],
            status=subscription_form.cleaned_data['status']
        )
        messages.info(request, _("Subscription Created"))

        return redirect(self.success_url)


class AdminSubscriptionAddPaymentView(LoginRequiredMixin, TemplateView):
    template_name = 'vendor/manage/subscription_add_payment.html'
    success_url = reverse_lazy('vendor_admin:manager-subscriptions')

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)

        subscription = Subscription.objects.get(uuid=kwargs.get('uuid_subscription'))
        profile = CustomerProfile.objects.get(uuid=kwargs.get('uuid_profile'))

        if 'site' in request.GET:
            offer_site_form = OfferSiteSelectForm(request.GET)

            if not offer_site_form.is_valid():
                context['offer_site_form'] = offer_site_form
                return render(request, self.template_name, context)
            
            context['form'] = SubscriptionAddPaymentForm(initial={
                'offer': offer_site_form.cleaned_data['offer'],
                'subscription': subscription,
                'profile': profile
            })
            
            context['offer_site_form'] = offer_site_form

            return render(request, self.template_name, context)

        context['offer_site_form'] = OfferSiteSelectForm()

        return render(request, self.template_name, context)

    def post(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        offer_site_form = OfferSiteSelectForm(request.GET)

        if not offer_site_form.is_valid():
            context['offer_site_form'] = payment_form
            return render(request, self.template_name, context)

        payment_form = SubscriptionAddPaymentForm(request.POST, site=offer_site_form.cleaned_data['site'])

        if not payment_form.is_valid():
            context['form'] = payment_form
            return render(request, self.template_name, context)
        
        payment = payment_form.save(commit=False)
        offer = payment_form.cleaned_data['offer']

        invoice = payment.profile.get_cart_or_checkout_cart()
        invoice.empty_cart()
        invoice.status = InvoiceStatus.COMPLETE
        invoice.add_offer(offer)

        try:
            with transaction.atomic():
                payment.invoice = invoice
                payment.amount = invoice.total
                payment.save()

                if payment.success and payment.status == PurchaseStatus.SETTLED:
                    receipt = Receipt.objects.create(
                        transaction=payment.transaction,
                        order_item=invoice.order_items.first(),
                        profile=payment.profile,
                        start_date=offer.get_offer_start_date(payment.submitted_date),
                        end_date=offer.get_offer_end_date(payment.submitted_date),
                        subscription=payment.subscription
                    )
                    receipt.products.add(offer.products.first())

            messages.info(request, _("Payment Added to Subscription"))

        except (IntegrityError, DatabaseError, Exception) as exce:
            logger.error(f"AdminSubscriptionCreateView error: {exce}")
            messages.error(request, "failed to add payment to subscription")
            return render(request, self.template_name, context)

        return redirect(request.META.get('HTTP_REFERER', self.success_url))


class AdminProfileListView(LoginRequiredMixin, TableFilterMixin, SiteOnRequestFilterMixin, ListView):
    """
    List of CustomerProfiles on site
    """
    template_name = "vendor/manage/profile_list.html"
    model = CustomerProfile
    paginate_by = 100

    def search_filter(self, queryset):
        search_value = self.request.GET.get('search_filter')
        return queryset.filter(Q(pk__icontains=search_value) | \
                               Q(user__email__icontains=search_value) | \
                               Q(user__username__icontains=search_value))

    def get_paginated_by(self, queryset):
        if 'paginate_by' in self.request.kwargs:
            return self.kwargs['paginate_by']
        return self.paginate_by
    
    def get_queryset(self):
        queryset = super().get_queryset()

        return queryset.order_by('pk')


class AdminProfileDetailView(LoginRequiredMixin, DetailView):
    '''
    Gets all Customer Profile information for quick lookup and management
    '''
    template_name = 'vendor/manage/profile_detail.html'
    model = CustomerProfile
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context['free_offers'] = Offer.objects.filter(prices__cost=0, site=self.object.site)
        context['invoices'] = self.object.invoices.order_by("-created")

        return context


class AdminManualSubscriptionRenewal(LoginRequiredMixin, DetailView):
    success_url = reverse_lazy('vendor_admin:manage-profiles')
    model = Subscription
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'

    def post(self, request, *args, **kwargs):
        subscription = Subscription.objects.get(uuid=self.kwargs["uuid"])
        submitted_datetime = timezone.now()

        invoice = Invoice.objects.create(
            profile=customer_profile,
            site=site,
            ordered_date=submitted_datetime,
            status=InvoiceStatus.COMPLETE
        )
        invoice.add_offer(offer)
        invoice.save()

        transaction_id = timezone.now().strftime("%Y-%m-%d_%H-%M-%S-Manual-Renewal")

        processor = get_site_payment_processor(invoice.site)(invoice.site, invoice)
        processor.renew_subscription(subscription, transaction_id, PurchaseStatus.CAPTURED, submitted_datetime)

        messages.info(request, _("Subscription Renewed"))
        return redirect(request.META.get('HTTP_REFERER', self.success_url))


class PaymentWithNoReceiptListView(LoginRequiredMixin, ListView):
    template_name = "vendor/manage/payment_list.html"
    model = Payment

    def get_queryset(self):
        site = get_site_from_request(self.request)
        return [payment for payment in Payment.objects.filter(invoice__site=site, success=True) if payment.get_receipt() is None]

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        context['title'] = _("Payments with no Receipts")
        return context


class PaymentWithNoOrderItemsListView(LoginRequiredMixin, ListView):
    template_name = "vendor/manage/payment_list.html"
    model = Payment

    def get_queryset(self):
        site = get_site_from_request(self.request)
        return Payment.objects.filter(invoice__site=site, success=True).annotate(order_item_count=Count('invoice__order_items')).filter(order_item_count=0)

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        context['title'] = _("Payments with no Order Items")
        return context



# App Settings
from django.conf import settings
from django import forms
from django.db.models import TextChoices
from django.contrib.sites.models import Site
from django.utils.translation import gettext_lazy as _

from siteconfigs.config import SiteConfigBaseClass


class SiteSelectForm(forms.Form):
    site = forms.ModelChoiceField(queryset=Site.objects.all())


class SupportedPaymentProcessor(TextChoices):
    PROMO_CODE_BASE = ("base.PaymentProcessorBase", _("Default Processor"))
    AUTHORIZE_NET = ("authorizenet.AuthorizeNetProcessor", _("Authorize.Net"))
    STRIPE = ("stripe.StripeProcessor", _("Stripe"))


class PaymentProcessorForm(SiteSelectForm):
    payment_processor = forms.CharField(
        label=_("Payment Processor"),
        widget=forms.Select(choices=SupportedPaymentProcessor.choices),
    )


class PaymentProcessorSiteConfig(SiteConfigBaseClass):
    label = _("Payment Processor")
    default = {"payment_processor": "base.PaymentProcessorBase"}
    form_class = PaymentProcessorForm
    key = ""
    instance = None

    def __init__(self, site=None):

        if site is None:
            site = Site.objects.get_current()

        self.key = ".".join([__name__, __class__.__name__])
        super().__init__(site, self.key)

        if not self.instance:
            self.default["payment_processor"] = VENDOR_PAYMENT_PROCESSOR

    def get_form(self):
        return self.form_class(initial=self.get_initials())

    def get_initials(self):
        initial = super().get_initials()
        initial["site"] = (self.site.pk, self.site.domain)

        if self.instance:
            initial["payment_processor"] = [
                choice
                for choice in SupportedPaymentProcessor.choices
                if choice[0] == self.instance.value["payment_processor"]
            ][0]
        else:
            initial["payment_processor"] = SupportedPaymentProcessor.choices[0]

        return initial

    def get_selected_processor(self):
        if self.instance:
            return [
                choice
                for choice in SupportedPaymentProcessor.choices
                if choice[0] == self.instance.value["payment_processor"]
            ][0]

        return SupportedPaymentProcessor.choices[0]  # Return Default Processors


class StripeConnectAccountForm(SiteSelectForm):
    account_number = forms.CharField(max_length=120)


class StripeConnectAccountConfig(SiteConfigBaseClass):
    label = _("Stripe Connect Account")
    default = {"stripe_connect_account": None}
    form_class = StripeConnectAccountForm
    key = ""
    instance = None

    def __init__(self, site=None):

        if site is None:
            site = Site.objects.get_current()

        self.key = ".".join([__name__, __class__.__name__])
        super().__init__(site, self.key)


class VendorSiteCommissionForm(SiteSelectForm):
    commission = forms.IntegerField(min_value=0, max_value=100)


class VendorSiteCommissionConfig(SiteConfigBaseClass):
    label = _("Vendor Site Commission")
    default = {"commission": None}
    form_class = VendorSiteCommissionForm
    key = ""
    instance = None

    def __init__(self, site=None):

        if site is None:
            site = Site.objects.get_current()

        self.key = ".".join([__name__, __class__.__name__])
        super().__init__(site, self.key)



VENDOR_PRODUCT_MODEL = getattr(settings, "VENDOR_PRODUCT_MODEL", "vendor.Product")

VENDOR_PAYMENT_PROCESSOR = getattr(
    settings, "VENDOR_PAYMENT_PROCESSOR", "dummy.DummyProcessor"
)

DEFAULT_CURRENCY = getattr(settings, "DEFAULT_CURRENCY", "usd")

AVAILABLE_CURRENCIES = getattr(
    settings, "AVAILABLE_CURRENCIES", {"usd": _("USD Dollars")}
)

VENDOR_STATE = getattr(settings, "VENDOR_STATE", "DEBUG")

# Encryption settings
VENDOR_DATA_ENCODER = getattr(
    settings, "VENDOR_DATA_ENCODER", "vendor.encrypt.cleartext"
)

ENABLE_STRIPE_SIGNALS = getattr(settings, 'ENABLE_STRIPE_SIGNALS', False)

STRIPE_BASE_COMMISSION = getattr(settings, 'STRIPE_BASE_COMMISSION', {'percentage': 2.9, 'fixed': 0.3})

STRIPE_RECURRING_COMMISSION = getattr(settings, 'STRIPE_RECURRING_COMMISSION', {'percentage': 0.5})



from calendar import monthrange
from datetime import datetime
from decimal import Decimal
from django import forms
from django.apps import apps
from django.conf import settings
from django.contrib.sites.models import Site
from django.forms import inlineformset_factory
from django.forms.widgets import SelectDateWidget, TextInput
from django.utils.translation import gettext_lazy as _

from integrations.models import Credential

from vendor.config import VENDOR_PRODUCT_MODEL
from vendor.models import (
    Address,
    Offer,
    Price,
    offer_term_details_default,
    CustomerProfile,
    Payment,
)
from vendor.models.choice import (
    PaymentTypes,
    RefundReasons,
    TermType,
    Country,
    USAStateChoices,
    SubscriptionStatus,
)
from vendor.utils import get_site_from_request
from vendor.config import SiteSelectForm


Product = apps.get_model(VENDOR_PRODUCT_MODEL)

COUNTRY_CHOICE = getattr(settings, 'VENDOR_COUNTRY_CHOICE', Country)
COUNTRY_DEFAULT = getattr(settings, 'VENDOR_COUNTRY_DEFAULT', Country.US)


def get_available_country_choices():
    if isinstance(COUNTRY_CHOICE, list):
        return [(country.value, country.label) for country in Country if country.name in COUNTRY_CHOICE]
    return Country.choices


class PriceForm(forms.ModelForm):
    CHOICES = [('not_free', _('Purchase Price')),
               ('free', _('Free'))]
    price_select = forms.ChoiceField(label="", choices=CHOICES, widget=forms.widgets.RadioSelect())

    class Meta:
        model = Price
        fields = ['price_select', 'cost', 'currency',
                  'start_date', 'end_date', 'priority']

    def __init__(self, *args, **kwargs):
        super(PriceForm, self).__init__(*args, **kwargs)
        self.fields['start_date'].widget.attrs['class'] = 'datepicker'
        self.fields['end_date'].widget.attrs['class'] = 'datepicker'
        self.fields['cost'].label = _('Price')
        self.fields['cost'].widget = TextInput()
        self.fields['cost'].widget.attrs['placeholder'] = '##.##'
        self.fields['cost'].widget.attrs['class'] = 'w-50'

        if self.instance.cost:
            self.initial['price_select'] = self.CHOICES[0]
        else:
            self.initial['price_select'] = self.CHOICES[1]


class ProductForm(forms.ModelForm):
    class Meta:
        model = Product
        fields = ['sku', 'name', 'site', 'available', 'description', 'meta']


class OfferForm(forms.ModelForm):
    products = forms.ModelMultipleChoiceField(label=_("Available Products:"), required=True, queryset=Product.objects.filter(site=settings.SITE_ID, available=True))

    class Meta:
        model = Offer
        fields = ['name', 'start_date', 'end_date', 'terms', 'term_details',
                  'term_start_date', 'available', 'offer_description', 'allow_multiple']

    def __init__(self, *args, **kwargs):
        request = kwargs.pop('request')
        super(OfferForm, self).__init__(*args, **kwargs)
        self.fields['start_date'].widget.attrs['class'] = 'datepicker'
        self.fields['end_date'].widget.attrs['class'] = 'datepicker'
        self.fields['term_start_date'].widget.attrs['class'] = 'datepicker'
        self.fields['available'].label = _('Available to Purchase')
        self.fields['products'].queryset = Product.objects.filter(site=get_site_from_request(request), available=True)

    def clean(self):
        cleaned_data = super().clean()

        if not cleaned_data['name']:
            product_names = [product.name for product in self.cleaned_data['products']]
            if len(product_names) == 1:
                self.cleaned_data['name'] = product_names[0]
            else:
                self.cleaned_data['name'] = _("Bundle: ") + ", ".join(product_names)

        term_detail_defaults = [False for default in list(offer_term_details_default().keys()) if default not in cleaned_data['term_details']]

        if self.data['terms'] == str(TermType.SUBSCRIPTION) and False in term_detail_defaults:
            self.add_error('term_details', _("Invalid term details for subscription"))

        return cleaned_data


class AddressForm(forms.ModelForm):

    class Meta:
        model = Address
        fields = ['name', 'first_name', 'last_name', 'country',
                  'address_1', 'address_2', 'locality', 'state', 'postal_code']

    def __init__(self, *args, **kwargs):
        super(AddressForm, self).__init__(*args, **kwargs)

        self.fields['name'].hidden = True
        self.fields['first_name'].widget.attrs.update({'placeholder': _('Enter First Name')})
        self.fields['last_name'].widget.attrs.update({'placeholder': _('Enter Last Name')})
        self.fields['country'].label = _("Billing Country/Region")

        self.fields['address_1'].widget.attrs.update({'placeholder': _('Enter Address')})
        self.fields['address_2'].widget.attrs.update({'placeholder': _('Enter Apt, Suite, Unit, Building, Floor, etc.')})
        self.fields['state'].widget.attrs.update({'placeholder': _('Enter State')})
        self.fields['locality'].widget.attrs.update({'placeholder': _('Enter City')})

        self.fields['postal_code'].required = True
        self.fields['postal_code'].label = _("Zip Code")
        self.fields['postal_code'].widget.attrs.update({'placeholder': _('Enter Zip')})
        self.fields['country'].choices = get_available_country_choices()
        
        if 'instance' in kwargs:
            self.initial['country'] = kwargs['instance'].country
        else:
            self.initial['country'] = Country[COUNTRY_DEFAULT].value


class AccountInformationForm(AddressForm):
    email = forms.EmailField(label=_('Email Address'), required=True)
    select_state = forms.ChoiceField(label=_("State"), choices=USAStateChoices.choices, required=False)
    prefix = "account_info"

    class Meta:
        model = Address
        fields = ['name', 'first_name', 'last_name', 'email', 'country',
                  'address_1', 'address_2', 'locality', 'state', 'postal_code']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['email'].widget.attrs.update({'placeholder': _('Enter Email Address')})
        self.fields['select_state'].choices = [("", _("Select State"))] + self.fields['select_state'].choices

    def clean(self):
        cleaned_data = super().clean()
        state = cleaned_data.get('state')
        select_state = cleaned_data.get('select_state')

        if not (state or select_state):
            self.add_error('select_state', _('This field is required.'))

        return cleaned_data


class BillingAddressForm(AddressForm):
    same_as_shipping = forms.BooleanField(label=_("Billing address is the same as shipping address"), required=False, initial=True)
    select_state = forms.ChoiceField(label=_("State"), choices=USAStateChoices.choices, required=False)
    prefix = "billing"

    class Meta:
        model = Address
        fields = ['same_as_shipping', 'name', 'first_name', 'last_name',
                  'country', 'address_1', 'address_2', 'locality', 'state', 'postal_code']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['select_state'].choices = [("", _("Select State"))] + self.fields['select_state'].choices

    def clean(self):
        cleaned_data = super().clean()
        state = cleaned_data.get('state')
        select_state = cleaned_data.get('select_state')

        if not (state or select_state):
            self.add_error('select_state', _('This field is required.'))

        return cleaned_data


class CreditCardField(forms.CharField):

    # validates almost all of the example cards from PayPal
    # https://www.paypalobjects.com/en_US/vhelp/paypalmanager_help/credit_card_numbers.htm
    cards = [
        {
            'type': 'maestro',
            'patterns': [5018, 502, 503, 506, 56, 58, 639, 6220, 67],
            'length': [12, 13, 14, 15, 16, 17, 18, 19],
            'cvvLength': [3],
            'luhn': True
        }, {
            'type': 'forbrugsforeningen',
            'patterns': [600],
            'length': [16],
            'cvvLength': [3],
            'luhn': True
        }, {
            'type': 'dankort',
            'patterns': [5019],
            'length': [16],
            'cvvLength': [3],
            'luhn': True
        }, {
            'type': 'visa',
            'patterns': [4],
            'length': [13, 16],
            'cvvLength': [3],
            'luhn': True
        }, {
            'type': 'mastercard',
            'patterns': [51, 52, 53, 54, 55, 22, 23, 24, 25, 26, 27],
            'length': [16],
            'cvvLength': [3],
            'luhn': True
        }, {
            'type': 'amex',
            'patterns': [34, 37],
            'length': [15],
            'cvvLength': [3, 4],
            'luhn': True
        }, {
            'type': 'dinersclub',
            'patterns': [30, 36, 38, 39],
            'length': [14],
            'cvvLength': [3],
            'luhn': True
        }, {
            'type': 'discover',
            'patterns': [60, 64, 65, 622],
            'length': [16],
            'cvvLength': [3],
            'luhn': True
        }, {
            'type': 'unionpay',
            'patterns': [62, 88],
            'length': [16, 17, 18, 19],
            'cvvLength': [3],
            'luhn': False
        }, {
            'type': 'jcb',
            'patterns': [35],
            'length': [16],
            'cvvLength': [3],
            'luhn': True
        }
    ]

    def __init__(self, placeholder=None, *args, **kwargs):
        super(CreditCardField, self).__init__(widget=forms.widgets.TextInput(attrs={'placeholder': placeholder, 'type': 'tel'}), *args, **kwargs)

    default_error_messages = {
        'invalid': _(u'The credit card number is invalid'),
    }

    def clean(self, value):

        # ensure no spaces or dashes
        value = value.replace(' ', '').replace('-', '')

        # get the card type and its specs
        card = self.card_from_number(value)

        # if no card found, invalid
        if not card:
            raise forms.ValidationError(self.error_messages['invalid'])

        # check the length
        if not len(value) in card['length']:
            raise forms.ValidationError(self.error_messages['invalid'])

        # test luhn if necessary
        if card['luhn']:
            if not self.validate_mod10(value):
                raise forms.ValidationError(self.error_messages['invalid'])

        return value

    def card_from_number(self, num):
        # find this card, based on the card number, in the defined set of cards
        for card in self.cards:
            for pattern in card['patterns']:
                if (str(pattern) == str(num)[:len(str(pattern))]):
                    return card

    def validate_mod10(self, num):
        # validate card number using the Luhn (mod 10) algorithm
        checksum, factor = 0, 1
        for x in reversed(num):
            for y in str(factor * int(x)):
                checksum += int(y)
            factor -= 3
            factor = abs(factor)
        return checksum % 10 == 0


class PaymentFrom(forms.Form):
    payment_type = forms.ChoiceField(label=_("Payment Type"), choices=PaymentTypes.choices, widget=forms.widgets.HiddenInput)


class PaymentRefundForm(forms.ModelForm):
    refund_amount = forms.DecimalField()
    reason = forms.ChoiceField(choices=RefundReasons.choices)
    void_end_date = forms.BooleanField(required=False)

    class Meta:
        model = Payment
        fields = ['refund_amount', 'reason', 'void_end_date']

    def clean_refund_amount(self):
        past_refunds_amount = 0
        refund_amount = self.cleaned_data.get('refund_amount', 0)

        if refund_amount > self.instance.amount:
            raise forms.ValidationError(_("Refund amount cannot be greater than the original amount"))
        
        if (past_refunds := self.instance.result.get('refunds', [])):
            for partial_refund in past_refunds:
                past_refunds_amount += Decimal(partial_refund.get('amount', 0))

        if (refund_amount + past_refunds_amount) > self.instance.amount:
            raise forms.ValidationError(_("Refund amount cannot be greater than the original amount"))

        return refund_amount


class CreditCardForm(PaymentFrom):
    full_name = forms.CharField(required=True, label=_("Name on Card"), max_length=80)
    card_number = CreditCardField(label=_("Credit Card Number"), placeholder=u'0000 0000 0000 0000', min_length=12, max_length=19)
    expire_month = forms.ChoiceField(required=True, label=_("Expiration Month"), choices=[(x, f'{x:02d}') for x in range(1, 13)])
    expire_year = forms.ChoiceField(required=True, label=_("Expiration Year"))
    cvv_number = forms.CharField(required=True, label=_("CVV Number"), max_length=4, min_length=3, widget=forms.TextInput(attrs={'size': '4', 'placeholder': _('Enter CVV')}))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        today = datetime.now()
        self.fields['expire_year'].choices = [(x, x) for x in range(today.year, today.year + 15)]
        self.fields['expire_year'].initial = (today.year, today.year)
        self.fields['full_name'].widget.attrs.update({'placeholder': _('Enter Name on Card')})

    def clean(self):
        cleaned_data = super(CreditCardForm, self).clean()
        expire_month = cleaned_data.get('expire_month')
        expire_year = cleaned_data.get('expire_year')

        if not self.expiration_date_valid(expire_month, expire_year):
            del(cleaned_data['expire_month'])
            del(cleaned_data['expire_year'])

        return cleaned_data

    def clean_expire_month(self):
        expire_month = self.cleaned_data.get('expire_month')

        if not expire_month:
            raise forms.ValidationError(_("You must select a valid expiration month"), code='required')

        return expire_month

    def clean_expire_year(self):
        expire_year = self.cleaned_data.get('expire_year')

        if not expire_year:
            raise forms.ValidationError(_("You must select a valid expiration year"), code='required')

        return expire_year

    def clean_cvv_number(self):
        cvv_number = self.cleaned_data.get('cvv_number')

        if not cvv_number.isdigit():
            raise forms.ValidationError(_("CVV must be all digits"))

        return cvv_number

    def expiration_date_valid(self, expire_month, expire_year):
        year = int(expire_year)
        month = int(expire_month)

        # find last day of the month
        day = monthrange(year, month)[1]
        expire = datetime(year, month, day)

        if datetime.now() > expire:
            self._errors["expire_year"] = self.error_class([_("The expiration date you entered is in the past.")])
            self._errors["expire_month"] = self.error_class([_("Check expiration")])
            return False
        return True


class DateRangeForm(forms.Form):
    start_date = forms.DateField(required=False, label=_("Start Date"), widget=SelectDateWidget())
    end_date = forms.DateField(required=False, label=_("End Date"), widget=SelectDateWidget())

    def clean(self):
        cleaned_data = super().clean()
        start_date = cleaned_data.get('start_date')
        end_date = cleaned_data.get('end_date')

        if end_date and end_date < start_date:
            self.add_error('end_date', _('End Date cannot be before Start Date'))
            del(cleaned_data['end_date'])

        return cleaned_data


class DateTimeRangeForm(forms.Form):
    start_date = forms.DateTimeField(required=False, label=_("Start Date"), widget=SelectDateWidget())
    end_date = forms.DateTimeField(required=False, label=_("End Date"), widget=SelectDateWidget())

    def clean(self):
        cleaned_data = super().clean()
        start_date = cleaned_data.get('start_date')
        end_date = cleaned_data.get('end_date')

        if end_date and end_date < start_date:
            self.add_error('end_date', _('End Date cannot be before Start Date'))
            del(cleaned_data['end_date'])

        return cleaned_data


class OfferSiteSelectForm(SiteSelectForm):
    offer = forms.ModelChoiceField(queryset=None, required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if 'site' in self.initial or 'site' in self.data:
            site = self.initial.get('site', self.data['site'])
            self.fields['offer'].queryset = Offer.objects.filter(site=site)
        else:
            self.fields['offer'].widget = forms.HiddenInput()


class SubscriptionForm(forms.Form):
    site = forms.ModelChoiceField(queryset=Site.objects.all())
    profile = forms.ModelChoiceField(queryset=None)
    subscription_id = forms.CharField(max_length=80)
    status = forms.ChoiceField(choices=SubscriptionStatus.choices)
        
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not ('site' in self.initial or 'site' in self.data):
            raise KeyError("site needs to be inluded in either initial or data attributes")

        site = self.initial.get('site', self.data.get('site'))
            
        self.fields['site'].widget = forms.HiddenInput()
        self.fields['profile'].queryset = CustomerProfile.objects.filter(site=site).order_by('user__username').select_related('user')


class SubscriptionAddPaymentForm(forms.ModelForm):
    offer = forms.ModelChoiceField(queryset=None)
    
    class Meta:
        model = Payment
        fields = ['subscription', 'profile', 'submitted_date', 'transaction', 'amount', 'success', 'status', 'payee_full_name']

    def __init__(self, *args, **kwargs):
        offers_queryset = None
        if 'site' in kwargs:
            offers_queryset = Offer.objects.filter(site=kwargs.get('site'))
            del(kwargs['site'])

        super().__init__(*args, **kwargs)

        self.fields['offer'].widget = forms.HiddenInput()
        self.fields['offer'].queryset = offers_queryset
        self.fields['submitted_date'].widget.attrs['class'] = 'datepicker'
        self.fields['subscription'].widget = forms.HiddenInput()
        self.fields['profile'].widget = forms.HiddenInput()


class StartDateForm(forms.Form):
    start_date = forms.DateTimeField(widget=forms.SelectDateWidget(attrs={"class": "datetime"}))

##########
# From Sets
##########


PriceFormSet = inlineformset_factory(
    Offer,
    Price,
    form=PriceForm,
    can_delete=False,
    exclude=('offer',),
    validate_max=True,
    min_num=1,
    extra=0)


class AuthorizeNetIntegrationForm(forms.ModelForm):

    class Meta:
        model = Credential
        fields = ['client_id', 'public_key', 'private_key']
        labels = {
            'client_id': _("API ID"),
            'public_key': _("Transaction Key"),
            'private_key': _("Signature ID")
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['client_id'].required = True
        self.fields['public_key'].required = True
        self.fields['private_key'].required = True


class StripeIntegrationForm(forms.ModelForm):

    class Meta:
        model = Credential
        fields = ['public_key', 'private_key']
        labels = {
            'public_key': _("Publishable Client Key"),
            'private_key': _("Secret Server Key")
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['public_key'].required = True
        self.fields['private_key'].required = True



import logging

from django.contrib import admin
from django.db.models import Count

from vendor.models import TaxClassifier, Offer, Price, CustomerProfile, \
    Invoice, OrderItem, Receipt, Wishlist, WishlistItem, Address, Payment, \
    Subscription
from vendor.models.choice import InvoiceStatus
from vendor.models.base import get_product_model

logger = logging.getLogger(__name__)


################
# ADMIN ACTIONS
################
def soft_delete_payments_with_no_receipt(modeladmin, request, queryset):
    '''
    This action finds any successfull payment that has not a receipt that matches its transaction and created fields, and soft deletes them.
    Every successfull payment has a receipt that has the same transaction value.
    '''
    invalid_payments = [payment for payment in Payment.objects.filter(success=True) if payment.get_receipt() is None]

    for payment in invalid_payments:
        payment.delete()

    logger.info(f"Soft deleted {len(invalid_payments)} payments")


def soft_delete_payments_without_order_items_invoice(modeladmin, request, queryset):
    '''
    This action finds any successfull payment that with an invoice that has no order_items.
    Meaning it is a useless payment as it did not pay for any item
    '''
    invalid_payments = Payment.objects.filter(success=True).annotate(order_item_count=Count('invoice__order_items')).filter(order_item_count=0)
    for payment in invalid_payments:
        payment.delete()

    logger.info(f"Soft deleted {len(invalid_payments)} payments")


def soft_delete_invoices_with_deleted_payments(modeladmin, request, queryset):
    '''
    This action finds any invoice that has a status greater then Checkout and has a soft deleted payment.
    If an invoice has two payments, both need to be deleted to delete the invoice.
    '''
    for invoice in Invoice.objects.filter(status__gt=InvoiceStatus.CHECKOUT):
        soft_delete = True
        for payment in invoice.payments.all():
            if not payment.deleted:
                soft_delete = False

        if soft_delete:
            invoice.delete()


soft_delete_payments_with_no_receipt.short_description = "Soft Delete Payments with no receipt"

soft_delete_payments_without_order_items_invoice.short_description = "Soft Delete Payments with zero order_items"

soft_delete_invoices_with_deleted_payments.short_description = "Soft Delete Invoices with payments that have been soft deleted"


###############
# INLINES
###############
class InvoiceInline(admin.TabularInline):
    model = Invoice
    extra = 1


class OrderItemInline(admin.TabularInline):
    model = OrderItem
    extra = 1


class PriceInline(admin.TabularInline):
    model = Price
    extra = 1


class ReceiptInline(admin.TabularInline):
    model = Receipt
    exclude = ('deleted', 'uuid', 'vendor_notes', 'meta', 'profile')
    readonly_fields = ('pk', 'transaction', 'order_item', 'start_date', 'end_date')
    max_num = 1


class PaymentInline(admin.TabularInline):
    model = Payment
    exclude = ('deleted', 'uuid', 'provider', 'profile', 'billing_address', 'result', 'payee_full_name', 'payee_company')
    readonly_fields = ('pk', 'transaction', 'invoice', 'created', 'status', 'amount', 'success')
    max_num = 1


class WishlistInline(admin.TabularInline):
    model = Wishlist
    extra = 1


class WishlistItemInline(admin.TabularInline):
    model = WishlistItem
    extra = 1


###############
# MODEL ADMINS
###############
class AddressAdmin(admin.ModelAdmin):
    readonly_fields = ('uuid',)
    list_display = ('name', 'profile', 'postal_code')
    search_fields = ('postal_code', )


class CustomerProfileAdmin(admin.ModelAdmin):
    readonly_fields = ('uuid', 'user')
    list_display = ('pk', 'user', 'email', 'site', 'currency', 'created')
    search_fields = ('user__username', 'user__email')
    list_filter = ('site__domain', )


class InvoiceAdmin(admin.ModelAdmin):
    readonly_fields = ('uuid', 'shipping_address', 'profile')
    list_display = ('pk', '__str__', 'profile', 'site', 'status', 'total', 'created', 'deleted')
    search_fields = ('uuid', 'profile__user__username', )
    list_filter = ('site__domain', )
    inlines = [
        OrderItemInline,
    ]
    actions = [soft_delete_invoices_with_deleted_payments]


class OfferAdmin(admin.ModelAdmin):
    readonly_fields = ('uuid',)
    list_display = ('name', 'slug', 'site', 'terms', 'available', 'deleted')
    search_fields = ('name', 'site__domain', )
    list_filter = ('site__domain', )
    inlines = [
        PriceInline,
    ]


class PaymentAdmin(admin.ModelAdmin):
    readonly_fields = ('uuid', 'invoice', 'profile' )
    list_display = ('pk', 'created', 'submitted_date', 'transaction', 'subscription', 'invoice', 'profile', 'amount', 'deleted', 'status')
    search_fields = ('pk', 'transaction', 'profile__user__username', )
    list_filter = ('profile__site__domain', 'success', 'status')
    exclude = ('billing_address', )
    actions = [soft_delete_payments_without_order_items_invoice, soft_delete_payments_with_no_receipt]


class ReceiptAdmin(admin.ModelAdmin):
    readonly_fields = ('uuid', 'order_item', 'profile', 'products')
    exclude = ('updated', )
    list_display = ('pk', 'transaction', 'subscription', 'created', 'profile', 'order_item', 'start_date', 'end_date', 'deleted')
    list_filter = ('profile__site__domain', 'products')
    search_fields = ('pk', 'transaction', 'profile__user__username', )
    # Example for future filters on AdminForm Fields
    # def get_form(self, request, obj=None, change=False, **kwargs):
    #     form = super().get_form(request, obj, change, **kwargs)

    #     if not obj:
    #         return form

    #     form.base_fields['profile'].queryset = CustomerProfile.objects.filter(site=obj.profile.site)

    #     return form


class SubscriptionAdmin(admin.ModelAdmin):
    readonly_fields = ('uuid', 'profile')
    exclude = ('updated', )
    list_display = ('pk', 'gateway_id', 'created', 'profile', 'status')
    list_filter = ('profile__site__domain', )
    search_fields = ('pk', 'gateway_id', 'profile__user__username', )
    inlines = [PaymentInline, ReceiptInline]


class PriceAdmin(admin.ModelAdmin):
    readonly_fields = ('offer', )
    list_display = ('pk', 'offer', 'cost')
    list_filter = ('offer', 'offer__site')
    search_fields = ('pk', 'cost', 'offer__name')


class TaxClassifierAdmin(admin.ModelAdmin):
    pass


class WishlistAdmin(admin.ModelAdmin):
    inlines = [
        WishlistItemInline,
    ]

###############
# REGISTRATION
###############


admin.site.register(TaxClassifier, TaxClassifierAdmin)
admin.site.register(CustomerProfile, CustomerProfileAdmin)
admin.site.register(Offer, OfferAdmin)
admin.site.register(Invoice, InvoiceAdmin)
admin.site.register(Wishlist, WishlistAdmin)
admin.site.register(Address, AddressAdmin)
admin.site.register(Receipt, ReceiptAdmin)
admin.site.register(Payment, PaymentAdmin)
admin.site.register(Subscription, SubscriptionAdmin)
admin.site.register(OrderItem)
admin.site.register(Price, PriceAdmin)



import hashlib
import hmac
import json
import logging

from datetime import datetime
from django.conf import settings
from django.core.exceptions import (MultipleObjectsReturned,
                                    ObjectDoesNotExist, PermissionDenied)
from django.db.models import Q
from django.http import HttpResponse, HttpResponseRedirect, JsonResponse
from django.urls import reverse_lazy
from django.utils import timezone
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.edit import FormMixin

from vendor.forms import DateTimeRangeForm
from vendor.integrations import AuthorizeNetIntegration
from vendor.models import Invoice, Payment, Subscription
from vendor.models.choice import InvoiceStatus, PurchaseStatus
from vendor.processors.authorizenet import (AuthorizeNetProcessor,
                                            sync_subscriptions)
from vendor.utils import get_site_from_request

logger = logging.getLogger(__name__)


def update_payment(site, transaction_id, transaction_detail):
    processor = AuthorizeNetProcessor(site)

    submitted_datetime = datetime.strptime(transaction_detail.submitTimeUTC.pyval, '%Y-%m-%dT%H:%M:%S.%f%z')
    payment_info = processor.get_payment_info(transaction_detail)
    payment_status = processor.get_payment_status(transaction_detail.transactionStatus.text)
    payment_success = processor.get_payment_success(transaction_detail.responseCode.text)

    try:
        payment = Payment.objects.get(profile__site=site, transaction=transaction_id, status__lt=PurchaseStatus.VOID)
        payment.submitted_date = submitted_datetime
        
        if not isinstance(payment.result, dict):
            payment.result = {}
            payment.result[timezone.now().strftime("%Y-%m-%d_%H:%M:%S")] = payment_info
        else:
            payment.result.update({
                timezone.now().strftime("%Y-%m-%d_%H:%M:%S"): payment_info
            })
        payment.status = payment_status
        payment.success = payment_success
        payment.save()
        
    except MultipleObjectsReturned as exce:
        logger.error(f"ERROR AuthorizeCaptureAPI update_payment multiple payments for transaction: {transaction_id} error: {exce}")

    except ObjectDoesNotExist as exce:
        logger.error(f"ERROR AuthorizeCaptureAPI update_payment payment does not exist for transaction: {transaction_id} error: {exce}")

    except Exception as exce:
        logger.error(f"ERROR AuthorizeCaptureAPI update_payment transaction: {transaction_id}, error: {exce}")


def subscription_save_transaction(site, transaction_id, transaction_detail):
    processor = AuthorizeNetProcessor(site)

    subscription_id = transaction_detail.subscription.id.text
    logger.info(f"AuthorizeCaptureAPI subscription_save_transaction saving subscription transaction: {transaction_id} for subscription {subscription_id}")

    submitted_datetime = datetime.strptime(transaction_detail.submitTimeUTC.pyval, '%Y-%m-%dT%H:%M:%S.%f%z')
    payment_info = processor.get_payment_info(transaction_detail)
    payment_status = processor.get_payment_status(transaction_detail.transactionStatus.text)
    payment_success = processor.get_payment_success(transaction_detail.responseCode.text)

    logger.info(f"AuthorizeCaptureAPI subscription_save_transaction detail: submitted_date: {submitted_datetime}, payment_info: {payment_info}, payment_status: {payment_status}, payment_success: {payment_success}")

    try:
        subscription = Subscription.objects.get(gateway_id=subscription_id, profile__site=site)

    except MultipleObjectsReturned as exce:
        logger.error(f"ERROR AuthorizeCaptureAPI subscription_save_transaction multiple subscription for id: {subscription_id} error: {exce}")
        subscription = Subscription.objects.filter(gateway_id=subscription_id, profile__site=site).first()

    except ObjectDoesNotExist as exce:
        logger.error(f"ERROR AuthorizeCaptureAPI subscription_save_transaction subscription does not exist {subscription_id} exce: {exce}")
        return None

    if not subscription.get_offer():
        logger.error(f"ERROR AuthorizeCaptureAPI subscription_save_transaction subscription does not have an Offer, subscription: {subscription.pk} - {subscription.gateway_id}")
        return None

    try:
        payment = subscription.payments.get(Q(transaction="") | Q(transaction=None), Q(status=PurchaseStatus.QUEUED))

    except MultipleObjectsReturned as exce:
        # There should be none or only one payment with transaction None and status in Queue
        logger.error(f"ERROR AuthorizeCaptureAPI MultipleObjectsReturned subscription_save_transaction multiple payments returned with None as Transaction, for {subscription_id} exce: {exce}")
        subscription.payments.filter(Q(transaction="") | Q(transaction=None), Q(status=PurchaseStatus.QUEUED)).delete()
        logger.info(f"AuthorizeCaptureAPI renew_subscription_task subscription {subscription.pk} renewed")
        invoice = Invoice.objects.create(
            profile=subscription.profile,
            site=site,
            ordered_date=submitted_datetime,
            total=transaction_detail.settleAmount.pyval,
            status=InvoiceStatus.COMPLETE
        )
        invoice.add_offer(subscription.get_offer())
        invoice.save()

        processor = AuthorizeNetProcessor(site, invoice)
        processor.subscription = subscription
        
        processor.renew_subscription(subscription, transaction_id, payment_status, payment_success, submitted_date=submitted_datetime)
        logger.info(f"AuthorizeCaptureAPI subscription_save_transaction creating new payment and receipt for subscription, for {subscription_id}")

        return None # No need to continue to create receipt as it is done in the above function

    except ObjectDoesNotExist as exce:
        logger.info(f"AuthorizeCaptureAPI ObjectDoesNotExist Payment for subscription: {subscription_id} - exce: {exce}")
        logger.info(f"AuthorizeCaptureAPI renew_subscription_task subscription {subscription.pk} renewed")
        invoice = Invoice.objects.create(
            profile=subscription.profile,
            site=site,
            ordered_date=submitted_datetime,
            total=transaction_detail.settleAmount.pyval,
            status=InvoiceStatus.COMPLETE
        )
        invoice.add_offer(subscription.get_offer())
        invoice.save()

        processor = AuthorizeNetProcessor(site, invoice)
        processor.subscription = subscription
        
        processor.renew_subscription(subscription, transaction_id, payment_status, payment_success, submitted_date=submitted_datetime)
        logger.info(f"AuthorizeCaptureAPI subscription_save_transaction creating new payment and receipt for subscription, for {subscription_id}")
        
        return None # No need to continue to create receipt as it is done in the above function
    
    payment.amount = transaction_detail.settleAmount.pyval
    payment.transaction = transaction_id
    payment.submitted_date = submitted_datetime
    payment.result = {}
    payment.result[timezone.now().strftime("%Y-%m-%d_%H:%M:%S")] = payment_info
    payment.status = payment_status
    payment.success = payment_success
    payment.save()
    logger.info(f"AuthorizeCaptureAPI subscription_save_transaction: payment {payment.pk} updated")
    
    if payment_status == PurchaseStatus.SETTLED:
        processor = AuthorizeNetProcessor(site, payment.invoice)
        processor.payment = payment
        processor.subscription = subscription
        processor.create_receipts(payment.invoice.order_items.all())
        logger.info(f"AuthorizeCaptureAPI subscription_save_transaction: subscription renewed {subscription.pk}")

def settle_authorizenet_transactions(site, start_date, end_date):
    processor = AuthorizeNetProcessor(site)

    settled_transactions = processor.get_settled_transactions(start_date, end_date)
    processor.update_payments_to_settled(site, settled_transactions)


class AuthorizeNetBaseAPI(View):
    """
    Base class to handel Authroize.Net webhooks.
    """

    @csrf_exempt
    def dispatch(self, *args, **kwargs):
        """
        Dispatch override to accept post entries without csfr given that they include
        a X-Anet-Signature in there header that must be encoded with Signature Key using
        HMAC-sha512 on the payload.
        Reference: https://developer.authorize.net/api/reference/features/webhooks.html
        """
        return super().dispatch(*args, **kwargs)

    def is_valid_post(self, site):
        logger.info(f"AuthorizeNetBaseAPI is_valid_post: Request body: {self.request.body}")
        logger.info(f"AuthorizeNetBaseAPI is_valid_post: X ANET SIGNATURE: {self.request.META.get('HTTP_X_ANET_SIGNATURE')}")

        if 'HTTP_X_ANET_SIGNATURE' not in self.request.META:
            logger.warning("AuthorizeNetBaseAPI is_valid_post: Signature Key not it request")
            return False

        try:
            credentials = AuthorizeNetIntegration(site)

            if credentials.instance and credentials.instance.private_key:
                hash_value = hmac.new(bytes(credentials.instance.private_key, 'utf-8'), self.request.body, hashlib.sha512).hexdigest()
            elif settings.AUTHORIZE_NET_SIGNATURE_KEY:
                hash_value = hmac.new(bytes(settings.AUTHORIZE_NET_SIGNATURE_KEY, 'utf-8'), self.request.body, hashlib.sha512).hexdigest()
            else:
                raise TypeError("No private key set")

            logger.info(f"AuthorizeNetBaseAPI is_valid_post: Checking hashs\nCALCULATED: {hash_value}\nREQUEST VALUE: {self.request.META.get('HTTP_X_ANET_SIGNATURE')[7:]}")

            if hash_value.upper() == self.request.META.get('HTTP_X_ANET_SIGNATURE')[7:]:
                return True

        except TypeError as exce:
            logger.error(f'ERROR AuthorizeNetBaseAPI is_valid_post: TypeError Exception: {exce}')

        return False


class AuthorizeCaptureAPI(AuthorizeNetBaseAPI):
    """
    API endpoint to get event notifications from authorizenet when a authcaputre is created.
    If there is a subscription tied to the transaction, it will renew such subscription
    """
    def is_request_valid(self, site, request_data):
        if not request_data:
            logger.warning("AuthorizeCaptureAPI post: Webhook event has no body")
            return False

        if not request_data.get('payload', {}).get('id'):
            logger.error(f"ERROR AuthorizeCaptureAPI post: No transaction id request data: {request_data}")
            return False
            
        if not self.is_valid_post(site):
            logger.error(f"ERROR AuthorizeCaptureAPI post: Request was denied: {self.request}")
            return False

        return True

    def post(self, *args, **kwargs):
        try:
            logger.info(f"AuthorizeCaptureAPI post: Event webhook: {self.request.body}")

            site = get_site_from_request(self.request)
            logger.info(f"AuthorizeCaptureAPI post: site: {site}")

            request_data = json.loads(self.request.body)
            logger.info(f"AuthorizeCaptureAPI post: request data: {request_data}")

            if not self.is_request_valid(site, request_data):
                logger.error("AuthorizeCaptureAPI post: Request is invalid")
                return HttpResponse(status=200, content="AuthorizeCaptureAPI post: Request is invalid")

            transaction_id = request_data.get('payload').get('id')
            logger.info(f"AuthorizeCaptureAPI post: Getting transaction detail for id: {transaction_id}")

            processor = AuthorizeNetProcessor(site)
            transaction_detail = processor.get_transaction_detail(transaction_id)

            if not hasattr(transaction_detail, 'subscription'):
                logger.info(f"AuthorizeCaptureAPI post: updating payment for transaction: {transaction_id}")
                update_payment(site, transaction_id, transaction_detail)

            elif transaction_detail:
                logger.info(f"AuthorizeCaptureAPI post: saving subscription transaction: {transaction_id}")
                subscription_save_transaction(site, transaction_id, transaction_detail)
            else:
                logger.error(f"ERROR AuthorizeCaptureAPI post: No transaction detail for transaction: {transaction_id}")

        except Exception as exce:
            logger.error(f"ERROR AuthorizeCaptureAPI post: exception: {exce}")
            
        logger.info("AuthorizeCaptureAPI post event finished")
        return HttpResponse(status=200, content="AuthorizeCaptureAPI post event finished")


class VoidAPI(AuthorizeNetBaseAPI):

    def post(self, *args, **kwargs):
        logger.info(f"VoidAPI post: Event webhook: {self.request.body}")
        site = get_site_from_request(self.request)
        logger.info(f"VoidAPI post: site: {site}")

        if not self.request.body:
            logger.warning("VoidAPI post: Webhook event has no body")
            return JsonResponse({"msg": "VoidAPI post: Webhook event has no body"})

        if not self.is_valid_post(site):
            logger.error(f"ERROR VoidAPI post: Request was denied: {self.request}")
            raise PermissionDenied()
        
        request_data = json.loads(self.request.body)
        logger.info(f"VoidAPI post: request data: {request_data}")

        if request_data.get('eventType') != 'net.authorize.payment.void.created':
            logger.error(f"ERROR VoidAPI post: wrong event type: {request_data.get('eventType')}")
            return JsonResponse({"msg": "Event type is incorrect"})
        
        Payment.objects.filter(profile__site=site, transaction=request_data.get('payload').get('id')).update(status=PurchaseStatus.VOID)
        logger.info(f"VoidAPI post: payment with transaction voided: {request_data.get('payload').get('id')}")

        return JsonResponse({"msg": "VoidAPI post: success"})


class SyncSubscriptionsView(View):

    def get(self, *args, **kwargs):
        site = get_site_from_request(self.request)
        sync_subscriptions(site)
        return JsonResponse({'msg': 'Sync Subscriptions Finished'})


class GetSettledTransactionsView(FormMixin, View):
    form_class = DateTimeRangeForm
    success_url = reverse_lazy('vendor:vendor-home')

    def post(self, request, *args, **kwargs):
        form = self.get_form_class()(request.POST)
        site = get_site_from_request(self.request)

        if form.is_valid():
            settle_authorizenet_transactions(site, form.cleaned_data['start_date'], form.cleaned_data['end_date'])

        return HttpResponseRedirect(self.request.META.get('HTTP_REFERER', self.get_success_url()))


from django.utils import timezone
from django.db.models import F
from django.shortcuts import get_object_or_404, render, redirect
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.urls import reverse, reverse_lazy
from django.conf import settings
from django.utils.translation import gettext as _
from django.core.exceptions import ObjectDoesNotExist

from django.views import View
from django.views.generic.edit import CreateView, UpdateView, DeleteView, FormView
from django.views.generic.list import ListView
from django.views.generic.detail import DetailView
from django.views.generic import TemplateView
from django.http import HttpResponse

from vendor.models import Offer, OrderItem, Invoice, Payment, Address
from vendor.models.address import Address as GoogleAddress
from vendor.models.choice import InvoiceStatus
from vendor.forms import BillingAddressForm, CreditCardForm

from .vendor_admin import AdminDashboardView, AdminInvoiceDetailView, AdminInvoiceListView
from .vendor import TransferExistingSubscriptionsToStripe

######################
# Order History Views
class OrderHistoryListView(LoginRequiredMixin, ListView):
    '''
    List of all the invoices generated by the current user on the current site.
    '''
    model = Invoice
    #TODO: filter to only include the current user's order history

    def get_queryset(self):
        try:
            return self.request.user.customer_profile.get().invoices.filter(status__gt=InvoiceStatus.CART)  # The profile and user are site specific so this should only return what's on the site for that user excluding the cart
        except ObjectDoesNotExist:         # Catch the actual error for the exception
            return []   # Return empty list if there is no customer_profile

class OrderHistoryDetailView(LoginRequiredMixin, DetailView):
    '''
    Details of an invoice generated by the current user on the current site.
    '''
    template_name = "vendor/invoice_history_detail.html"
    model = Invoice
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'


# Generated by Django 3.1.3 on 2021-06-02 15:35

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


def set_default_user():
    """
    Function will return the value for MIGRATION_0023_DEFAULT_USER declared in the
    the settings.py file. If the constant is not declared or is None it will return
    1 to be set as a default value.
    """
    if hasattr(settings, 'MIGRATION_0023_DEFAULT_USER') and settings.MIGRATION_0023_DEFAULT_USER is not None:
        return settings.MIGRATION_0023_DEFAULT_USER
    return 1


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('vendor', '0022_offer_meta'),
    ]

    operations = [
        migrations.AlterField(
            model_name='customerprofile',
            name='user',
            field=models.ForeignKey(default=set_default_user, on_delete=django.db.models.deletion.CASCADE, related_name='customer_profile', to=settings.AUTH_USER_MODEL, verbose_name='User'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='invoice',
            name='profile',
            field=models.ForeignKey(default=set_default_user, on_delete=django.db.models.deletion.CASCADE, related_name='invoices', to='vendor.customerprofile', verbose_name='Customer Profile'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='payment',
            name='profile',
            field=models.ForeignKey(blank=True, default=set_default_user, on_delete=django.db.models.deletion.CASCADE, related_name='payments', to='vendor.customerprofile', verbose_name='Purchase Profile'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='receipt',
            name='profile',
            field=models.ForeignKey(default=set_default_user, on_delete=django.db.models.deletion.CASCADE, related_name='receipts', to='vendor.customerprofile', verbose_name='Purchase Profile'),
            preserve_default=False,
        ),
    ]
